from __future__ import print_function, unicode_literals, absolute_import, division

# non-standard imports
from astropy import units as u
from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.coordinates import CartesianRepresentation
from astropy.utils import lazyproperty
from scipy.interpolate import interp1d
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import path, transforms, patches, colors
from matplotlib.collections import PatchCollection

# optional
try:
    from ginga import trcalc
    from ginga.util import wcs
    import ginga.canvas.types.all as ginga_types

    has_ginga = True
except Exception:
    has_ginga = False

try:
    from importlib import resources as importlib_resources
except Exception:
    # backport for Python 3.6
    import importlib_resources

flip = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])

# COMPO is only available on GTC, so these values are well-known
pixel_scale = 0.08086 * u.arcsec / u.pix
focal_plane_scale = 1.214 * u.arcsec / u.mm
gtc_focalplane_equivalencies = [
    (u.mm, u.arcsec, lambda x: x * 1.214, lambda x: x / 1.214),
    (u.pix, u.arcsec, lambda x: x / 0.08086, lambda x: x * 0.08086),
    (u.mm, u.pix, lambda x: x * 1.214 / 0.08086, lambda x: x * 0.08086 / 1.214),
]

# predicted position of pick-off pupil from FoV centre
# from Zeemax simulations
THETA = u.Quantity([0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65], unit=u.deg)
X = u.Quantity(
    [
        0.0,
        0.476,
        0.949,
        1.414,
        1.868,
        2.309,
        2.731,
        3.133,
        3.511,
        3.863,
        4.185,
        4.475,
        4.731,
        4.951,
    ],
    unit=u.arcmin,
)
Y = u.Quantity(
    [
        0.0,
        0.021,
        0.083,
        0.186,
        0.329,
        0.512,
        0.732,
        0.988,
        1.278,
        1.600,
        1.951,
        2.329,
        2.731,
        3.154,
    ],
    unit=u.arcmin,
)

# Vital statistics for COMPO (from FDR and commissioning)
PARK_POSITION = -66.4 * u.deg
MAX_ANGLE = 55 * u.deg
PICKOFF_SIZE = 26.73 * u.arcsec  # 330 pixels
MIRROR_SIZE = 24.3 * u.arcsec  # 20 mm
SHADOW_X = 40 * u.mm  # extent of vignetting by injector arm
SHADOW_Y = 49 * u.mm  # extent of vignetting by injector arm (~739 pix)
INJECTOR_THETA = 13.04 * u.deg  # angle of injector arm when in position
GUIDE_THETA = -66.4 * u.deg  # angle of injector arm when guiding
LENS_REF_POSITION = 0 * u.mm  # TODO: this is made up, replace with true value
NOMINAL_INJECTOR_ZERO = -0.1583 * u.deg  # value of stage encoder when arm centred
NOMINAL_PICKOFF_ZERO = 0.34 * u.deg  # value of stage encoder when arm centred
MAX_ANGLE = 67 * u.deg  # hard stops at (67.79, -67.55) for PO, (67.4, -67.8) for IA

# interpolated functions for X and Y positions - not unit aware
x_func = interp1d(THETA, X, kind="cubic", bounds_error=False, fill_value="extrapolate")
y_func = interp1d(THETA, Y, kind="cubic", bounds_error=False, fill_value="extrapolate")

# interpolated functions for lens offset
try:
    lens_data_file = (
        importlib_resources.files("hcam_widgets") / "data" / "compo_lens_offset.csv"
    )
    _, po_theta, lens_off = np.loadtxt(lens_data_file, delimiter=",", skiprows=1).T
except AttributeError:
    # Python < P3.9 fall back to deprecated pkg_resources
    from pkg_resources import resource_filename

    fn = resource_filename("hcam_widgets", "data/compo_lens_offset.csv")
    _, po_theta, lens_off = np.loadtxt(fn, delimiter=",", skiprows=1).T

_g = interp1d(po_theta, lens_off, bounds_error=False, fill_value="extrapolate")


@u.quantity_input(pickoff_theta=u.deg)
def target_lens_position(pickoff_theta, guiding=False):
    """
    Find the correct position for the corrector lens
    """
    # TODO: find correct value for guiding!
    offset = _g(abs(pickoff_theta.to_value(u.deg))) * u.mm
    return LENS_REF_POSITION + offset


@u.quantity_input(theta=u.deg)
def field_stop_centre(theta):
    """
    Where does the field stop of the injector arm fall in the focal plane?

    This is based on Zeemax simulations that include distortions
    and the curvature of the plane.
    """
    # make a simple float or numpy array
    theta = theta.to_value(u.deg)

    # flip the sign of the x-position if we have a -ve angle
    xmul = xmul = np.where(theta < 0, -1.0, 1.0)

    # interpolate
    theta = np.fabs(theta)
    return xmul * x_func(theta) * u.arcmin, y_func(theta) * u.arcmin


def focal_plane_units(unit):
    return unit in (u.deg, u.arcmin, u.arcsec, u.rad)


def focal_plane_to_sky(cartrep):
    """
    Convert physical position in focal plane to sky offset from FoV
    """
    cartrep = cartrep.transform(flip)
    with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
        return CartesianRepresentation(cartrep.xyz.to(u.arcsec))


@u.quantity_input(pickoff_angle=u.deg)
@u.quantity_input(injection_angle=u.deg)
def plot_compo(pickoff_angle, injection_angle, axis=None):
    if axis is None:
        fig, axis = plt.subplots()
    c = Chip().to_patches()
    poa = PickoffArm().to_patches(pickoff_angle)
    ia = InjectionArm().to_patches(injection_angle)
    # todo set colors with pc.set_array
    cmap = colors.ListedColormap(
        [
            "#BFD7EA",  # Chip (blue)
            "#59C9A5",  # POA arc
            "#56E39F",  # POA Baffle
            "#3A3042",  # POA FOV
            "#FF8C42",  # Inj Baffle
            "#3A3042",  # Inj Fov
        ]
    )
    pc = PatchCollection(c + poa + ia, alpha=0.8, cmap=cmap)
    pc.set_array(np.array([0, 1, 2, 3, 4, 5]))
    axis.add_collection(pc)
    axis.set_xlim(-300, 300)
    axis.set_ylim(-50, 150)
    axis.set_aspect("equal")
    return axis


class Chip:
    """
    Representing the Chip
    """

    with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
        NX = (2048 * u.pix).to(u.mm)
        NY = (1024 * u.pix).to(u.mm)

    @lazyproperty
    def vertices(self):
        return CartesianRepresentation(
            [-self.NX / 2, self.NX / 2, self.NX / 2, -self.NX / 2],
            [-self.NY / 2, -self.NY / 2, self.NY / 2, self.NY / 2],
            0 * u.mm,
        )

    def to_patches(self, unit=u.mm):
        with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
            ll = ((-self.NX / 2).to_value(unit), (-self.NY / 2).to_value(unit))
            width = self.NX.to_value(unit)
            height = self.NY.to_value(unit)
            rect = patches.Rectangle(ll, width, height)
        return [rect]

    def contains(self, repr):
        with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
            x = repr.x.to(u.mm)
            y = repr.y.to(u.mm)
        inchip = False
        if (-self.NX / 2 < x < self.NX / 2) and (-self.NY / 2 < y < self.NY / 2):
            inchip = True
        return inchip

    def clip_shape(self, vertices):
        """
        Clip a shape defined as a CartesianRepresentation of points by the chip edges

        Works in 2D - i.e drops the z-axis - as this is a projected clipping

        Notes
        -----
        This makes use of Sutherland-Hodgman clipping as implemented in
        AGG 2.4 and exposed in Matplotlib.

        Parameters
        ----------
        vertices :  astropy.coordinates.CartesianRepresentation
            vertices of shape to be clipped

        Returns
        --------
        astropy.coordinates.CartesianRepresentation
            new vertices, after clipping
        """
        # enforce pixel space
        try:
            with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
                xyz = vertices.xyz.to(u.mm)
        except u.UnitConversionError:
            raise ValueError("vertices are not in units of physical length")

        # drop Z axis and reshape to (N, 2)
        xy = xyz[:2].T
        poly = path.Path(xy, closed=True)
        clip_rect = transforms.Bbox(
            [
                [-self.NX.value / 2, -self.NY.value / 2],
                [self.NX.value / 2, self.NY.value / 2],
            ]
        )
        poly_clipped = poly.clip_to_bbox(clip_rect).to_polygons()[0]
        if np.all(poly_clipped[0] == poly_clipped[-1]):
            poly_clipped = poly_clipped[:-1]

        return CartesianRepresentation(
            *u.Quantity(poly_clipped, unit=u.mm).T, vertices.z
        )


class Baffle:
    """
    The Baffle present on both arms
    """

    BAFFLE_X = SHADOW_X
    BAFFLE_Y = SHADOW_Y
    INJECTION_ROTATION_RIGHT = rotation_matrix(-INJECTOR_THETA, "z")
    INJECTION_ROTATION_LEFT = rotation_matrix(INJECTOR_THETA, "z")
    INJECTION_TRANSLATION_RIGHT = CartesianRepresentation(
        57.5 * u.mm, 23.2 * u.mm, 0 * u.mm
    )
    INJECTION_TRANSLATION_LEFT = CartesianRepresentation(
        -57.5 * u.mm, 23.2 * u.mm, 0 * u.mm
    )

    @lazyproperty
    def vertices(self):
        return CartesianRepresentation(
            [
                -self.BAFFLE_X / 2,
                self.BAFFLE_X / 2,
                self.BAFFLE_X / 2,
                -self.BAFFLE_X / 2,
            ],
            [
                -self.BAFFLE_Y / 2,
                -self.BAFFLE_Y / 2,
                self.BAFFLE_Y / 2,
                self.BAFFLE_Y / 2,
            ],
            0 * u.mm,
        )


class InjectionArm:
    @u.quantity_input(theta=u.deg)
    def position(self, theta):
        """
        Position of pupil stop centre in focal plane (mm)

        Uses approximate formula which agrees with Zeemax calculations inc distortion to 1pix
        """
        r = 254.72 * u.mm  # length of injection arm
        off = 270 * u.mm  # dist from rotation axis to FoV centre
        d = off - r
        x = r * np.sin(-theta)
        y = d + r * (1 - np.cos(-theta))
        return CartesianRepresentation(x, y, 0 * u.mm)

    @u.quantity_input(theta=u.deg)
    def to_patches(self, theta, unit=u.mm):
        with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
            centre = self.position(theta)
            fov = patches.Circle(
                tuple(centre.xyz[:2].to_value(unit)),
                radius=MIRROR_SIZE.to_value(unit) / 2,
            )

            baffle_cart = Baffle().vertices
            baffle_cart = baffle_cart.transform(rotation_matrix(theta))
            baffle_cart += centre
            c = Chip()
            if c.contains(centre):
                baffle_cart = c.clip_shape(baffle_cart)
            baffle_xy = baffle_cart.xyz[:2].T.to_value(unit)
            baffle = patches.Polygon(baffle_xy, closed=True)

        return [baffle, fov]

    @u.quantity_input(theta=u.deg)
    @u.quantity_input(ra_cen=u.deg)
    @u.quantity_input(dec_cen=u.deg)
    def to_ginga_object(self, theta, ra_cen, dec_cen, **params):
        if not has_ginga:
            raise RuntimeError("ginga not installed, cannot create Ginga object")

        with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
            # centre w.r.t FoV
            centre = self.position(theta)

            # now Baffle
            baffle_cart = Baffle().vertices
            baffle_cart = baffle_cart.transform(rotation_matrix(theta))
            baffle_cart += centre
            c = Chip()
            if c.contains(centre):
                baffle_cart = c.clip_shape(baffle_cart)

            # convert to sky plane
            centre = focal_plane_to_sky(centre)
            baffle_cart = focal_plane_to_sky(baffle_cart)

            # add centre of pointing
            x, y = wcs.add_offset_radec(
                ra_cen.to_value(u.deg),
                dec_cen.to_value(u.deg),
                *centre.xyz[:2].to_value(u.deg)
            )
            fov = ginga_types.Circle(
                x, y, MIRROR_SIZE.to_value(u.deg) / 2, coord="wcs", **params
            )
            # add vignetting by baffle
            corners = [
                wcs.add_offset_radec(
                    ra_cen.to_value(u.deg),
                    dec_cen.to_value(u.deg),
                    *p.xyz[:2].to_value(u.deg)
                )
                for p in baffle_cart
            ]
            vignetting = ginga_types.Polygon(corners, coord="wcs", **params)
            return ginga_types.CompoundObject(vignetting, fov)


class PickoffArm:
    @u.quantity_input(theta=u.deg)
    def position(self, theta):
        """
        Position of pupil stop centre in focal plane (mm)

        Uses approximate formula which agrees with Zeemax calculations inc distortion to 1pix
        """
        r = 270 * u.mm
        # stage for pickoff arm rotates in opposite sense to coordinate axes
        x = r * np.sin(theta)
        y = r * (1 - np.cos(theta))
        # actually not in focal plane, but assuming it is is OK
        return CartesianRepresentation(x, y, 0 * u.mm)

    @u.quantity_input(theta=u.deg)
    def to_patches(self, theta, unit=u.mm):
        with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
            centre = self.position(theta)
            patrol_arc_centre = (0, (270 * u.mm).to_value(unit))
            radius = (270 * u.mm + 0.5 * MIRROR_SIZE).to_value(unit)
            arc = patches.Wedge(
                patrol_arc_centre,
                radius,
                # mpl, 0 is along x axis. COMPO, 0 is along -y
                PARK_POSITION.to_value(u.deg) - 90,
                MAX_ANGLE.to_value(u.deg) - 90,
                width=MIRROR_SIZE.to_value(unit),
            )
            pickoff = patches.Circle(
                tuple(centre.xyz[:2].to_value(unit)),
                radius=MIRROR_SIZE.to_value(unit) / 2,
            )
            baffle_cart = Baffle().vertices
            baffle_cart = baffle_cart.transform(rotation_matrix(-theta))
            baffle_cart += centre
            baffle_xy = baffle_cart.xyz[:2].T.to_value(unit)
            baffle = patches.Polygon(baffle_xy, closed=True)

        return [arc, baffle, pickoff]

    @u.quantity_input(theta=u.deg)
    @u.quantity_input(ra_cen=u.deg)
    @u.quantity_input(dec_cen=u.deg)
    def to_ginga_object(self, theta, ra_cen, dec_cen, **params):
        if not has_ginga:
            raise RuntimeError("ginga not installed, cannot create Ginga object")

        with u.set_enabled_equivalencies(gtc_focalplane_equivalencies):
            # centre w.r.t FoV
            centre = self.position(theta)

            # now Baffle
            baffle_cart = Baffle().vertices
            baffle_cart = baffle_cart.transform(rotation_matrix(-theta))
            baffle_cart += centre

            # convert to sky plane
            centre = focal_plane_to_sky(centre)
            baffle_cart = focal_plane_to_sky(baffle_cart)

            # create Pickoff circle
            x, y = wcs.add_offset_radec(
                ra_cen.to_value(u.deg),
                dec_cen.to_value(u.deg),
                *centre.xyz[:2].to_value(u.deg)
            )
            fov = ginga_types.Circle(
                x, y, MIRROR_SIZE.to_value(u.deg) / 2, coord="wcs", **params
            )
            # create vignetting by baffle
            corners = [
                wcs.add_offset_radec(
                    ra_cen.to_value(u.deg),
                    dec_cen.to_value(u.deg),
                    *p.xyz[:2].to_value(u.deg)
                )
                for p in baffle_cart
            ]
            vignetting = ginga_types.Polygon(corners, coord="wcs", **params)

            # TODO patrol arc.
            return ginga_types.CompoundObject(vignetting, fov)



from __future__ import print_function, unicode_literals, absolute_import, division
import logging
import time
import six
import sys

from .tkutils import get_root
if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


class GuiHandler(logging.Handler):
    """
    This defines the output sent to a text widget GUI
    """
    def __init__(self, twidget):
        """
        twidget : text widget to display logging messages
        """
        logging.Handler.__init__(self)
        logging.Formatter.converter = time.gmtime
        formatter = logging.Formatter('%(asctime)s - %(message)s\n', '%H:%M:%S')
        self.setFormatter(formatter)

        # ignore DEBUG messages
        self.setLevel(logging.INFO)

        # configure and store the text widget
        g = get_root(twidget).globals
        twidget.tag_config('DEBUG', background=g.COL['debug'])
        twidget.tag_config('INFO')
        twidget.tag_config('WARNING', background=g.COL['warn'])
        twidget.tag_config('ERROR', background=g.COL['error'])
        twidget.tag_config('CRITICAL', background=g.COL['critical'])
        self.twidget = twidget

    def emit(self, message):
        formattedMessage = self.format(message)

        # Write message to twidget
        g = get_root(self.twidget).globals
        self.twidget.configure(state=tk.NORMAL, font=g.ENTRY_FONT)
        self.twidget.insert(tk.END, formattedMessage, (message.levelname,))

        # Prevent further input
        self.twidget.configure(state=tk.DISABLED)
        self.twidget.see(tk.END)


class FileHandler(logging.FileHandler):
    """
    Used to send logging output to a file
    """
    def __init__(self, fname):
        """
        fout: file pointer to send messages to
        """
        logging.FileHandler.__init__(self, fname)
        logging.Formatter.converter = time.gmtime
        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)-7s %(message)s',
                                      '%Y-%m-%d %H:%M:%S')
        self.setFormatter(formatter)

        # include DEBUG messages
        self.setLevel(logging.DEBUG)


class StreamHandler(logging.StreamHandler):
    """
    Used to send logging output to stderr
    """
    def __init__(self):
        """
        fout: file pointer to send messages to
        """
        logging.StreamHandler.__init__(self)
        logging.Formatter.converter = time.gmtime
        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)-7s %(message)s\n',
                                      '%H:%M:%S')
        self.setFormatter(formatter)
        self.setLevel(logging.INFO)

    def emit(self, message):
        """
        Overwrites the default handler's emit method:

        message : the message to display
        """
        sys.stderr.write(self.format(message))


class Logger(object):
    """
    Defines an object for logging. This uses the logging module to
    define an internal logger and then defines how it reports to
    stderr and, optionally, a file. It also defines shortcuts to
    the standard logging methods warn, info etc. This is mainly
    as a base class for two GUI-based loggers that come next.

     logname : unique name for logger

    """

    def __init__(self, logname):

        # make a logger
        self._log = logging.getLogger(logname)

        # disable automatic logging to the terminal
        self._log.propagate = False

        # add terminal handler that avoids debug messages
        self._log.addHandler(StreamHandler())

    def update(self, fname):
        """
        Adds a handler to save to a file. Includes debug stuff.
        """
        ltfh = FileHandler(fname)
        self._log.addHandler(ltfh)

    def debug(self, message):
        self._log.debug(message)

    def info(self, message):
        self._log.info(message)

    def warn(self, message):
        self._log.warn(message)

    def error(self, message):
        self._log.error(message)

    def critical(self, message):
        self._log.critical(message)



#!/usr/bin/env python
"""
ucam provides classes and data specific to ULTRACAM
"""
from __future__ import print_function, absolute_import, unicode_literals, division
import six
import math

# internal imports
from . import widgets as w
from .tkutils import get_root

if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


# Timing, gain, noise parameters
# gains in electrons per count
GAIN_FAST = 1.4
GAIN_SLOW = 1.3
GAIN_TURBO = 1.5

# readout noise in electrons for 1x1, 2x2, 4x4, 8x8
READ_NOISE_TURBO = [7.0, 7.0, 7.0, 7.0]
READ_NOISE_FAST = [4.9, 4.9, 5.1, 6.4]
READ_NOISE_SLOW = [3.6, 3.6, 4.0, 5.4]
DARK_COUNT = 0.1  # counts/sec/pixel

# timing parameters in microseconds
INVERSION_DELAY = 110.
VCLOCK_FRAME = 23.3
VCLOCK_STORAGE = 23.3
HCLOCK = 0.48
CDS_TIME_FDD = 1.84
CDS_TIME_FBB = 4.4
CDS_TIME_CDD = 10.
SWITCH_TIME = 1.2


class InstPars(tk.LabelFrame):
    """
    Ultracam instrument parameters block.
    """

    def __init__(self, master):
        tk.LabelFrame.__init__(self, master, text='Instrument setup', padx=10, pady=10)

        # left hand side
        lhs = tk.Frame(self)
        # Application (mode)
        tk.Label(lhs, text='Mode').grid(row=0, column=0, sticky=tk.W)
        self.app = w.Radio(lhs, ('FF', 'Wins', 'Drift'), 3, self.check,
                           ('FullFrame', 'Windows', 'Drift'))
        self.app.grid(row=0, column=1, sticky=tk.W)

        # Clear enabled
        self.clearLab = tk.Label(lhs, text='Clear')
        self.clearLab.grid(row=1, column=0, sticky=tk.W)
        self.clear = w.OnOff(lhs, True, self.check)
        self.clear.grid(row=1, column=1, sticky=tk.W)

        # o/scan enabled
        self.oscanLab = tk.Label(lhs, text='O/Scan')
        self.oscanLab.grid(row=2, column=0, sticky=tk.W)
        self.oscan = w.OnOff(lhs, False, self.check)
        self.oscan.grid(row=2, column=1, sticky=tk.W)

        # Readout speed
        tk.Label(lhs, text='Readout speed').grid(row=3, column=0, sticky=tk.NW)
        self.readSpeed = w.Radio(lhs, ('Slow', 'Fast', 'Turbo'), 1,
                                 self.check, ('Slow', 'Fast', 'Turbo'))
        self.readSpeed.grid(row=3, column=1, pady=2, sticky=tk.W)

        # Exposure delay
        tk.Label(lhs, text='Exp. delay (ms)').grid(row=4, column=0,
                                                   sticky=tk.W)
        # exposure delay
        self.expose = w.PosInt(lhs, 0, None, True, width=6)
        self.expose.grid(row=4, column=1, columnspan=5, sticky=tk.W, pady=5)

        # Right-hand side: the window parameters
        rhs = tk.Frame(self)
        # window parameters
        xsl = (100, 100, 100)
        xslmin = (1, 1, 1)
        xslmax = (512, 512, 512)
        xsr = (600, 600, 600)
        xsrmin = (513, 513, 513)
        xsrmax = (1024, 1024, 1024)
        ys = (1, 201, 401)
        ysmin = (1, 1, 1)
        ysmax = (1024, 1024, 1024)
        nx = (50, 50, 50)
        ny = (50, 50, 50)
        xbfac = (1, 2, 3, 4, 5, 6, 8)
        ybfac = (1, 2, 3, 4, 5, 6, 8)
        self.wframe = w.WinPairs(rhs, xsl, xslmin, xslmax, xsr, xsrmin,
                                 xsrmax, ys, ysmin, ysmax, nx, ny,
                                 xbfac, ybfac, self.check, hcam=False)
        self.wframe.grid(row=2, column=0, columnspan=3, sticky=tk.W, pady=5)

        # Pack two halfs
        lhs.pack(side=tk.LEFT, anchor=tk.N, padx=5)
        rhs.pack(side=tk.LEFT, anchor=tk.N, padx=5)

        self.check()

    @property
    def isFF(self):
        return self.app.value() == 'FullFrame'

    @property
    def isDriftMode(self):
        return self.app.value() == 'Drift'

    def check(self, *args):
        """
        Checks the validity of the parameters.
        """
        g = get_root(self).globals

        status = True
        if self.isDriftMode:
            self.wframe.npair.set(1)

        if self.isFF:
            self.wframe.disable()
        else:
            self.wframe.enable()

        # can only have overscan with FF mode
        if self.oscan() and not self.isFF:
            g.clog.error('can only enable overscan in FF mode')
            status = False
            self.oscan.set(False)

        if self.app.value == "Windows" and self.wframe.npair.value() > 1 and self.clear():
            g.clog.error('clear mode only allowed with 1 pair')
            status = False
            self.clear.set(False)

        if status and hasattr(g, 'count'):
            # if valid, update timing and SN info
            if hasattr(g.count, 'update'):
                g.count.update()
            else:
                print('no count update')

        # check windows
        return self.wframe.check() and status

    def timing(self):
        """
        Estimates timing information for the current setup. You should
        run a check on the instrument parameters before calling this.

        Returns: (expTime, deadTime, cycleTime, dutyCycle)

        expTime   : exposure time per frame (seconds)
        deadTime  : dead time per frame (seconds)
        cycleTime : sampling time (cadence), (seconds)
        dutyCycle : percentage time exposing.
        frameRate : number of frames per second
        """
        # video timing
        readSpeed = self.readSpeed.value()
        if readSpeed == 'Fast':
            cdsTime = CDS_TIME_FBB
        elif readSpeed == 'Turbo':
            cdsTime = CDS_TIME_FDD
        elif readSpeed == 'Slow':
            cdsTime = CDS_TIME_CDD
        else:
            raise RuntimeError('unknown read speed')
        video = cdsTime + SWITCH_TIME

        # exposure delay in microsecs
        expose = self.expose.value() * 1000

        # binning values
        xbin = self.wframe.xbin.value()
        ybin = self.wframe.ybin.value()

        # check all the mode cases
        if self.isFF and not self.oscan():
            frameTransfer = 1033*VCLOCK_FRAME
            readout = (VCLOCK_STORAGE*ybin + 536*HCLOCK + (512/xbin+2)*video)*(1024/ybin)
            if self.clear():
                clearTime = (1033 + 1027)*VCLOCK_STORAGE
                cycleTime = (INVERSION_DELAY + expose + clearTime + frameTransfer + readout)/1.e6
                exposureTime = expose/1e6
            else:
                cycleTime = (INVERSION_DELAY + expose + frameTransfer + readout)/1.e6
                exposureTime = cycleTime - frameTransfer/1.0e6
            readout /= 1.0e6

        elif self.isFF and self.oscan():
            clearTime = (1033 + 1032) * VCLOCK_FRAME
            frameTransfer = 1033*VCLOCK_FRAME
            readout = (VCLOCK_STORAGE*ybin + 540*HCLOCK + ((540/xbin)+2)*video)*(1032/ybin)
            cycleTime = (INVERSION_DELAY + expose + clearTime + frameTransfer + readout)/1.e6
            exposureTime = expose/1e6
            readout /= 1.e6

        elif self.isDriftMode:
            xsl, xsr, ys, nx, ny = self.wframe.params(0)
            # number of windows stacked in storage
            nwins = int(0.5*(1.0 + 1033/ny))
            # pipe shift
            pshift = int(1033 - (2*nwins-1)*ny)

            frameTransfer = (ny + ys - 1)*VCLOCK_FRAME
            diffshift = abs(xsl - 1 - (1034-xsr-nx+1))

            if xsl - 1 > 1024 - xsr - nx + 1:
                numHclocks = nx + diffshift + (1024 - xsr - nx + 1) + 8
            else:
                numHclocks = nx + diffshift + (xsl - 1) + 8
            lineread = VCLOCK_STORAGE*ybin + numHclocks*HCLOCK + (nx/xbin+2)*video
            read = lineread * ny/ybin

            cycleTime = (INVERSION_DELAY + pshift*VCLOCK_STORAGE +
                         expose + frameTransfer + read)/1.e6
            exposureTime = cycleTime - frameTransfer/1.e6
            readout = (read + pshift*VCLOCK_STORAGE)/1.e6
        else:
            # windowed mode
            clearTime = (1033 + 1027)*VCLOCK_STORAGE if self.clear() else 0
            frameTransfer = 1033.*VCLOCK_FRAME
            cycleTime = INVERSION_DELAY + expose + frameTransfer + clearTime
            readout = 0.0
            # loop over all pairs
            for i in range(self.wframe.npair.value()):
                xsl, xsr, ys, nx, ny = self.wframe.params(i)

                # y params of previous window
                if i > 0:
                    ystart_m = self.wframe.ys[i-1].value()
                    ny_m = self.wframe.ny[i-1].value()
                else:
                    ystart_m = 1
                    ny_m = 0

                # Time taken to shift the window next to the storage area
                yshift = (ys-ystart_m-ny_m)*VCLOCK_STORAGE

                # Number of columns to shift whichever window is further from
                # the edge of the readout to get ready for simultaneous readout.
                diffshift = abs(xsl - 1 - (1024 - xsr - nx + 1))

                """
                Time taken to dump any pixels in a row that come after the ones we want.
                The '8' is the number of HCLOCKs needed to open the serial register dump gates
                If the left window is further from the left edge than the right window is from the
                right edge, then the diffshift will move it to be the same as the right window, and
                so we use the right window parameters to determine the number of hclocks needed, and
                vice versa.
                """
                if xsl - 1 > 1024 - xsr - nx + 1:
                    numHclocks = nx + diffshift + (1024 - xsr - nx + 1) + 8
                else:
                    numHclocks = nx + diffshift + (xsl - 1) + 8

                # Time taken to read one line.
                # The extra 2 is required to fill the video pipeline buffer
                lineRead = VCLOCK_STORAGE*ybin + numHclocks*HCLOCK + (nx/xbin+2)*video

                # time to read the window
                read = (ny/ybin)*lineRead

                cycleTime += read + yshift
                readout += read + yshift

            if self.clear():
                exposureTime = expose/1e6
            else:
                exposureTime = (cycleTime - frameTransfer)/1e6
            cycleTime /= 1e6

        deadTime = cycleTime - exposureTime
        frameRate = 1./cycleTime
        dutyCycle = 100.0*exposureTime/cycleTime
        return exposureTime, deadTime, cycleTime, dutyCycle, frameRate


class CountsFrame(tk.LabelFrame):
    """
    Frame for count rate estimates
    """
    def __init__(self, master):
        """
        master : enclosing widget
        """
        tk.LabelFrame.__init__(
            self, master, pady=2, text='Count & S-to-N estimator')

        # divide into left and right frames
        lframe = tk.Frame(self, padx=2)
        rframe = tk.Frame(self, padx=2)

        # entries
        self.filter = w.Radio(
            lframe, ('u', 'g', 'r', 'i', 'z'), 3, self.checkUpdate, initial=1)
        self.mag = w.RangedFloat(
            lframe, 18., 0., 30., self.checkUpdate, True, width=5)
        self.seeing = w.RangedFloat(
            lframe, 1.0, 0.2, 20., self.checkUpdate, True, True, width=5)
        self.airmass = w.RangedFloat(
            lframe, 1.5, 1.0, 5.0, self.checkUpdate, True, width=5)
        self.moon = w.Radio(lframe, ('d', 'g', 'b'),  3, self.checkUpdate)

        # results
        self.cadence = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.exposure = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.duty = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.peak = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.total = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.ston = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.ston3 = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)

        # layout
        # left
        tk.Label(lframe, text='Filter:').grid(
            row=0, column=0, padx=5, pady=3, sticky=tk.W+tk.N)
        self.filter.grid(row=0, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text='Mag:').grid(
            row=1, column=0, padx=5, pady=3, sticky=tk.W)
        self.mag.grid(row=1, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text='Seeing:').grid(
            row=2, column=0, padx=5, pady=3, sticky=tk.W)
        self.seeing.grid(row=2, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text='Airmass:').grid(
            row=3, column=0, padx=5, pady=3, sticky=tk.W)
        self.airmass.grid(row=3, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text='Moon:').grid(
            row=4, column=0, padx=5, pady=3, sticky=tk.W)
        self.moon.grid(row=4, column=1, padx=5, pady=3, sticky=tk.W)

        # right
        tk.Label(rframe, text='Cadence:').grid(
            row=0, column=0, padx=5, pady=3, sticky=tk.W)
        self.cadence.grid(row=0, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='Exposure:').grid(
            row=1, column=0, padx=5, pady=3, sticky=tk.W)
        self.exposure.grid(row=1, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='Duty cycle:').grid(
            row=2, column=0, padx=5, pady=3, sticky=tk.W)
        self.duty.grid(row=2, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='Peak:').grid(
            row=3, column=0, padx=5, pady=3, sticky=tk.W)
        self.peak.grid(row=3, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='Total:').grid(
            row=4, column=0, padx=5, pady=3, sticky=tk.W)
        self.total.grid(row=4, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='S/N:').grid(
            row=5, column=0, padx=5, pady=3, sticky=tk.W)
        self.ston.grid(row=5, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='S/N (3h):').grid(
            row=6, column=0, padx=5, pady=3, sticky=tk.W)
        self.ston3.grid(row=6, column=1, padx=5, pady=3, sticky=tk.W)

        # slot frames in
        lframe.grid(row=0, column=0, sticky=tk.W+tk.N)
        rframe.grid(row=0, column=1, sticky=tk.W+tk.N)

    def checkUpdate(self, *args):
        """
        Updates values after first checking instrument parameters are OK.
        This is not integrated within update to prevent ifinite recursion
        since update gets called from ipars.
        """
        g = get_root(self).globals
        if not self.check():
            g.clog.error('Current observing parameters are not valid.')
            return False

        if not g.ipars.check():
            g.clog.error('Current instrument parameters are not valid.')
            return False

    def check(self):
        """
        Checks values
        """
        status = True
        g = get_root(self).globals
        if self.mag.ok():
            self.mag.config(bg=g.COL['main'])
        else:
            self.mag.config(bg=g.COL['warn'])
            status = False

        if self.airmass.ok():
            self.airmass.config(bg=g.COL['main'])
        else:
            self.airmass.config(bg=g.COL['warn'])
            status = False

        if self.seeing.ok():
            self.seeing.config(bg=g.COL['main'])
        else:
            self.seeing.config(bg=g.COL['warn'])
            status = False

        return status

    def update(self, *args):
        """
        Updates values. You should run a check on the instrument and
        target parameters before calling this.
        """
        g = get_root(self).globals
        expTime, deadTime, cycleTime, dutyCycle, frameRate = g.ipars.timing()

        total, peak, peakSat, peakWarn, ston, ston3 = \
            self.counts(expTime, cycleTime)

        if cycleTime < 0.01:
            self.cadence.config(text='{0:7.5f} s'.format(cycleTime))
        elif cycleTime < 0.1:
            self.cadence.config(text='{0:6.4f} s'.format(cycleTime))
        elif cycleTime < 1.:
            self.cadence.config(text='{0:5.3f} s'.format(cycleTime))
        elif cycleTime < 10.:
            self.cadence.config(text='{0:4.2f} s'.format(cycleTime))
        elif cycleTime < 100.:
            self.cadence.config(text='{0:4.1f} s'.format(cycleTime))
        elif cycleTime < 1000.:
            self.cadence.config(text='{0:4.0f} s'.format(cycleTime))
        else:
            self.cadence.config(text='{0:5.0f} s'.format(cycleTime))

        if expTime < 0.01:
            self.exposure.config(text='{0:7.5f} s'.format(expTime))
        elif expTime < 0.1:
            self.exposure.config(text='{0:6.4f} s'.format(expTime))
        elif expTime < 1.:
            self.exposure.config(text='{0:5.3f} s'.format(expTime))
        elif expTime < 10.:
            self.exposure.config(text='{0:4.2f} s'.format(expTime))
        elif expTime < 100.:
            self.exposure.config(text='{0:4.1f} s'.format(expTime))
        elif expTime < 1000.:
            self.exposure.config(text='{0:4.0f} s'.format(expTime))
        else:
            self.exposure.config(text='{0:5.0f} s'.format(expTime))

        self.duty.config(text='{0:4.1f} %'.format(dutyCycle))
        self.peak.config(text='{0:d} cts'.format(int(round(peak))))
        if peakSat:
            self.peak.config(bg=g.COL['error'])
        elif peakWarn:
            self.peak.config(bg=g.COL['warn'])
        else:
            self.peak.config(bg=g.COL['main'])

        self.total.config(text='{0:d} cts'.format(int(round(total))))
        self.ston.config(text='{0:.1f}'.format(ston))
        self.ston3.config(text='{0:.1f}'.format(ston3))

    def counts(self, expTime, cycleTime, ap_scale=1.6, ndiv=5):
        """
        Computes counts per pixel, total counts, sky counts
        etc given current magnitude, seeing etc. You should
        run a check on the instrument parameters before calling
        this.

        expTime   : exposure time per frame (seconds)
        cycleTime : sampling, cadence (seconds)
        ap_scale  : aperture radius as multiple of seeing

        Returns: (total, peak, peakSat, peakWarn, ston, ston3)

        total    -- total number of object counts in aperture
        peak     -- peak counts in a pixel
        peakSat  -- flag to indicate saturation
        peakWarn -- flag to indication level approaching saturation
        ston     -- signal-to-noise per exposure
        ston3    -- signal-to-noise after 3 hours on target
        """
        # code directly translated from Java equivalent.
        g = get_root(self).globals

        # Set the readout speed
        readSpeed = g.ipars.readSpeed.value()
        xbin, ybin = g.ipars.wframe.xbin.value(), g.ipars.wframe.ybin.value()
        # index for readout noise array
        if xbin == 1:
            read_idx = 0
        elif xbin < 4:
            read_idx = 1
        elif xbin < 4:
            read_idx = 2
        else:
            read_idx = 3

        # calculate SN info.
        zero, sky, skyTot, gain, read, darkTot = 0., 0., 0., 0., 0., 0.
        total, peak, correct, signal, readTot, seeing = 0., 0., 0., 0., 0., 0.
        noise, _, narcsec, npix, signalToNoise3 = 1., 0., 0., 0., 0.

        tinfo = g.TINS[g.cpars['telins_name']]
        filtnam = self.filter.value()
        zero = tinfo['zerop'][filtnam]
        mag = self.mag.value()
        seeing = self.seeing.value()
        sky = g.SKY[self.moon.value()][filtnam]
        airmass = self.airmass.value()

        # GAIN, RNO
        if readSpeed == 'Fast':
            gain = GAIN_FAST
            read = READ_NOISE_FAST[read_idx]
        elif readSpeed == 'Turbo':
            gain = GAIN_TURBO
            read = READ_NOISE_TURBO[read_idx]
        elif readSpeed == 'Slow':
            gain = GAIN_SLOW
            read = READ_NOISE_SLOW[read_idx]

        plateScale = tinfo['plateScale']

        # calculate expected electrons
        total = 10.**((zero-mag-airmass*g.EXTINCTION[filtnam])/2.5)*expTime

        # compute fraction that fall in central pixel
        # assuming target exactly at its centre. Do this
        # by splitting each pixel of the central (potentially
        # binned) pixel into ndiv * ndiv points at
        # which the seeing profile is added. sigma is the
        # RMS seeing in terms of pixels.
        sigma = seeing/g.EFAC/plateScale

        sum = 0.
        for iyp in range(ybin):
            yoff = -ybin/2.+iyp
            for ixp in range(xbin):
                xoff = -xbin/2.+ixp
                for iys in range(ndiv):
                    y = (yoff + (iys+0.5)/ndiv)/sigma
                    for ixs in range(ndiv):
                        x = (xoff + (ixs+0.5)/ndiv)/sigma
                        sum += math.exp(-(x*x+y*y)/2.)
        peak = total*sum/(2.*math.pi*sigma**2*ndiv**2)

        # Work out fraction of flux in aperture with radius AP_SCALE*seeing
        correct = 1. - math.exp(-(g.EFAC*ap_scale)**2/2.)

        # expected sky e- per arcsec
        skyPerArcsec = 10.**((zero-sky)/2.5)*expTime
        # skyPerPixel = skyPerArcsec*plateScale**2*xbin*ybin
        narcsec = math.pi*(ap_scale*seeing)**2
        skyTot = skyPerArcsec*narcsec
        npix = math.pi*(ap_scale*seeing/plateScale)**2/xbin/ybin

        signal = correct*total        # in electrons
        darkTot = npix*DARK_COUNT*expTime  # in electrons
        readTot = npix*read**2         # in electrons

        # noise, in electrons
        noise = math.sqrt(readTot + darkTot + skyTot + signal)

        # Now compute signal-to-noise in 3 hour seconds run
        signalToNoise3 = signal/noise*math.sqrt(3*3600./cycleTime)

        # if using the avalanche mode, check that the signal level
        # is safe. A single electron entering the avalanche register
        # results in a distribution of electrons at the output with
        # mean value given by the parameter avalanche_gain. The
        # distribution is close to exponential, hence the probability
        # of obtaining an amplification n times higher than the mean is
        # given by e**-n. A value of 3/5 for n is adopted here for
        # warning/safety, which will occur once in every ~20/100
        # amplifications

        # convert from electrons to counts
        total /= gain
        peak /= gain

        warn = 25000
        sat = 60000

        peakSat = peak > sat
        peakWarn = peak > warn

        return (total, peak, peakSat, peakWarn, signal/noise, signalToNoise3)



# -*- coding: utf-8 -*-
"""
uspec provides classes and data specific to ULTRASPEC
"""
from __future__ import print_function, absolute_import, unicode_literals, division
import six
import math

# internal imports
from .tkutils import get_root
from . import widgets as w

if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk

# Timing, gain, noise parameters lifted from java usdriver
VCLOCK = 14.4e-6  # vertical clocking time
HCLOCK_NORM = 0.48e-6  # normal mode horizontal clock
HCLOCK_AV = 0.96e-6  # avalanche mode horizontal clock
VIDEO_NORM_SLOW = 11.20e-6
VIDEO_NORM_MED = 6.24e-6
VIDEO_NORM_FAST = 3.20e-6
VIDEO_AV_SLOW = 11.20e-6
VIDEO_AV_MED = 6.24e-6
VIDEO_AV_FAST = 3.20e-6
FFX = 1072
FFY = 1072
IFY = 1072
IFX = 1072
AVALANCHE_PIXELS = 1072
AVALANCHE_GAIN_9 = 1200.0  # dimensionless gain, hvgain=9
AVALANCHE_SATURATE = 80000   # electrons

# avalanche gains assume HVGain = 9. We can adapt this later when we decide
# how gain should be set at TNO. Might be better to make gain a function if
# we allow 0 < HVgain < 9 (SL)

GAIN_NORM_FAST = 0.8   # electrons per count
GAIN_NORM_MED = 0.7    # electrons per count
GAIN_NORM_SLOW = 0.8   # electrons per count
GAIN_AV_FAST = 0.0034  # electrons per count
GAIN_AV_MED = 0.0013   # electrons per count
GAIN_AV_SLOW = 0.0016  # electrons per count

# Note - avalanche RNO assume HVGain = 9. We can adapt this later when we
# decide how gain should be set at TNO. Might be better to make RNO a function
# if we allow 0 < HVgain < 9 (SL)

RNO_NORM_FAST = 4.8  # electrons per pixel
RNO_NORM_MED = 2.8   # electrons per pixel
RNO_NORM_SLOW = 2.2  # electrons per pixel
RNO_AV_FAST = 6.5    # electrons per pixel
RNO_AV_MED = 7.8     # electrons per pixel
RNO_AV_SLOW = 5.6    # electrons per pixel

# other noise sources
DARK_E = 0.001  # electrons/pix/sec
CIC = 0.010     # Clock induced charge, electrons/pix


class InstPars(tk.LabelFrame):
    """
    Ultraspec instrument parameters block.
    """

    def __init__(self, master):
        """
        master : enclosing widget
        """
        tk.LabelFrame.__init__(self, master, text='Instrument parameters',
                               padx=10, pady=10)

        # left hand side
        lhs = tk.Frame(self)

        # Application (mode)
        tk.Label(lhs, text='Mode').grid(row=0, column=0, sticky=tk.W)
        self.app = w.Radio(lhs, ('Wins', 'Drift'), 2, self.check,
                           ('Windows', 'Drift'))
        self.app.grid(row=0, column=1, sticky=tk.W)

        # Clear enabled
        self.clearLab = tk.Label(lhs, text='Clear')
        self.clearLab.grid(row=1, column=0, sticky=tk.W)
        self.clear = w.OnOff(lhs, True, self.check)
        self.clear.grid(row=1, column=1, sticky=tk.W)

        # Avalanche settings
        tk.Label(lhs, text='Avalanche').grid(row=2, column=0, sticky=tk.W)
        aframe = tk.Frame(lhs)
        self.avalanche = w.OnOff(aframe, False, self.check)
        self.avalanche.pack(side=tk.LEFT)
        self.avgainLabel = tk.Label(aframe, text='gain ')
        self.avgainLabel.pack(side=tk.LEFT)
        self.avgain = w.RangedInt(aframe, 0, 0, 9, self.check,
                                  False, width=2)
        self.avgain.pack(side=tk.LEFT)
        aframe.grid(row=2, column=1, pady=2, sticky=tk.W)

        # Readout speed
        tk.Label(lhs, text='Readout speed').grid(row=3, column=0, sticky=tk.NW)
        self.readSpeed = w.Radio(lhs, ('Slow', 'Medium', 'Fast'), 1,
                                 self.check, ('Slow', 'Medium', 'Fast'))
        self.readSpeed.grid(row=3, column=1, pady=2, sticky=tk.W)

        # Exposure delay
        tk.Label(lhs, text='Exposure delay (s)').grid(row=4, column=0,
                                                      sticky=tk.W)

        g = get_root(self).globals
        elevel = g.cpars['expert_level']
        if elevel == 0:
            self.expose = w.Expose(lhs, 0.0007, 0.0007, 1677.7207,
                                   self.check, width=7)
        elif elevel == 1:
            self.expose = w.Expose(lhs, 0.0007, 0.0003, 1677.7207,
                                   self.check, width=7)
        else:
            self.expose = w.Expose(lhs, 0.0007, 0., 1677.7207,
                                   self.check, width=7)
        self.expose.grid(row=4, column=1, pady=2, sticky=tk.W)

        # Number of exposures
        tk.Label(lhs, text='Num. exposures  ').grid(row=5, column=0, sticky=tk.W)
        self.number = w.PosInt(lhs, 1, None, False, width=7)
        self.number.grid(row=5, column=1, pady=2, sticky=tk.W)

        # LED setting
        self.ledLab = tk.Label(lhs, text='LED setting')
        self.ledLab.grid(row=6, column=0, sticky=tk.W)
        self.led = w.RangedInt(lhs, 0, 0, 4095, None, False, width=7)
        self.led.grid(row=6, column=1, pady=2, sticky=tk.W)
        self.ledValue = self.led.value()

        # Right-hand side: the window parameters
        rhs = tk.Frame(self)

        # window mode frame (initially full frame)
        xs = (1, 101, 201, 301)
        xsmin = (1, 1, 1, 1)
        xsmax = (1056, 1056, 1056, 1056)
        ys = (1, 101, 201, 301)
        ysmin = (1, 1, 1, 1)
        ysmax = (1072, 1072, 1072, 1072)
        nx = (1056, 100, 100, 100)
        ny = (1072, 100, 100, 100)
        xbfac = (1, 2, 3, 4, 5, 6, 8)
        ybfac = (1, 2, 3, 4, 5, 6, 8)
        self.wframe = w.Windows(rhs, xs, xsmin, xsmax, ys, ysmin, ysmax,
                                nx, ny, xbfac, ybfac, self.check)
        self.wframe.grid(row=2, column=0, columnspan=3, sticky=tk.W+tk.N)

        # drift mode frame (just one pair)
        xsl = (100,)
        xslmin = (1,)
        xslmax = (1024,)
        xsr = (600,)
        xsrmin = (1,)
        xsrmax = (1024,)
        ys = (1,)
        ysmin = (1,)
        ysmax = (1024,)
        nx = (50,)
        ny = (50,)
        xbfac = (1, 2, 3, 4, 5, 6, 8)
        ybfac = (1, 2, 3, 4, 5, 6, 8)
        self.pframe = w.WinPairs(rhs, xsl, xslmin, xslmax, xsr, xsrmin,
                                 xsrmax, ys, ysmin, ysmax, nx, ny,
                                 xbfac, ybfac, self.check, hcam=False)

        # Pack two halfs
        lhs.pack(side=tk.LEFT, anchor=tk.N, padx=5)
        rhs.pack(side=tk.LEFT, anchor=tk.N, padx=5)

        # Store freeze state
        self.frozen = False

        # stores current avalanche setting to check for changes
        self.oldAvalanche = False

        self.setExpertLevel()

    def setExpertLevel(self):
        """
        Modifies widget according to expertise level, which in this
        case is just matter of hiding or revealing the LED option
        and changing the lower limit on the exposure button.
        """
        g = get_root(self).globals
        level = g.cpars['expert_level']

        if level == 0:
            self.expose.fmin = 0.0007
            self.ledLab.grid_forget()
            self.led.grid_forget()
            self.ledValue = self.led.value()
            self.led.set(0)

        elif level == 1:
            self.expose.fmin = 0.0003
            self.led.set(self.ledValue)
            self.ledLab.grid(row=6, column=0, sticky=tk.W)
            self.led.grid(row=6, column=1, pady=2, sticky=tk.W)

        elif level == 2:
            self.expose.fmin = 0.0
            self.led.set(self.ledValue)
            self.ledLab.grid(row=6, column=0, sticky=tk.W)
            self.led.grid(row=6, column=1, pady=2, sticky=tk.W)

    def isDrift(self):
        """
        Returns True if we are in drift mode
        """
        if self.app.value() == 'Drift':
            return True
        elif self.app.value() == 'Windows':
            return False
        else:
            raise ValueError('uspec.InstPars.isDrift: application = ' +
                             self.app.value() + ' not recognised.')

    def loadXML(self, xml):
        """
        Sets the values of instrument parameters given an
        ElementTree containing suitable XML
        """
        g = get_root(self).globals
        # find application
        xmlid = xml.attrib['id']
        for app, d in g.cpars['templates'].iteritems():
            if xmlid == d['id']:
                break
        else:
            raise ValueError('Do not recognize application id = ' + xmlid)

        # find parameters
        cconfig = xml.find('configure_camera')
        pdict = {}
        for param in cconfig.findall('set_parameter'):
            pdict[param.attrib['ref']] = param.attrib['value']

        xbin, ybin = int(pdict['X_BIN']), int(pdict['Y_BIN'])

        # Set them.

        # Number of exposures
        self.number.set(pdict['NUM_EXPS'] if pdict['NUM_EXPS'] != '-1' else 0)

        # LED level
        self.led.set(pdict['LED_FLSH'])

        # Avalanche or normal
        self.avalanche.set(pdict['OUTPUT'])

        # Avalanche gain
        self.avgain.set(pdict['HV_GAIN'])

        # Dwell
        self.expose.set(str(float(pdict['DWELL'])/10000.))

        # Readout speed
        speed = pdict['SPEED']
        self.readSpeed.set('Slow' if
                           speed == '0' else 'Medium' if speed == '1'
                           else 'Fast')

        if app == 'Windows':
            # Clear or not
            self.clear.set(pdict['EN_CLR'])

            # now for the windows which come in two flavours
            self.app.set('Windows')
            w = self.wframe

            # X-binning factor
            w.xbin.set(xbin)

            # Y-binning factor
            w.ybin.set(ybin)

            # Load up windows
            nwin = 0
            for nw in range(4):
                xs = 'X' + str(nw+1) + '_START'
                ys = 'Y' + str(nw+1) + '_START'
                nx = 'X' + str(nw+1) + '_SIZE'
                ny = 'Y' + str(nw+1) + '_SIZE'
                if xs in pdict and ys in pdict and nx in pdict and ny in pdict \
                        and pdict[nx] != '0' and pdict[ny] != 0:
                    xsv, ysv, nxv, nyv = int(pdict[xs]), int(pdict[ys]), int(pdict[nx]), int(pdict[ny])
                    nxv *= xbin
                    nyv *= ybin

                    nchop = max(0, 17-xsv)
                    if nchop % xbin != 0:
                        nchop = xbin * (nchop // xbin + 1)

                    if self.avalanche():
                        xsv = max(1, 1074 - xsv - nxv)
                    else:
                        xsv = max(1, xsv + nchop - 16)
                    nxv -= nchop

                    w.xs[nw].set(xsv)
                    w.ys[nw].set(ysv)
                    w.nx[nw].set(nxv)
                    w.ny[nw].set(nyv)
                    nwin += 1
                else:
                    break

            # Set the number of windows
            w.nwin.set(nwin)

        else:
            self.clear.set(0)

            # now for drift mode
            self.app.set('Drift')
            p = self.pframe

            # X-binning factor
            p.xbin.set(xbin)

            # Y-binning factor
            p.ybin.set(ybin)

            # Load up window pair values
            xslv, xsrv, ysv, nxv, nyv = (
                int(pdict['X1_START']), int(pdict['X2_START']),
                int(pdict['Y1_START']), int(pdict['X1_SIZE']), int(pdict['Y1_SIZE'])
            )
            nxv *= xbin
            nyv *= ybin

            nchop = max(0, 17-xslv)
            if nchop % xbin != 0:
                nchop = xbin * (nchop // xbin + 1)

            if self.avalanche():
                xslv = max(1, 1074-xslv-nxv)
                xsrv = max(1, 1074-xsrv-nxv)
            else:
                xslv = max(1, xslv+nchop-16)
                xsrv = max(1, xsrv+nchop-16)

            nxv -= nchop
            if xslv > xsrv:
                xsrv, xslv = xslv, xsrv

            # finally set the values
            p.xsl[0].set(xslv)
            p.xsr[0].set(xsrv)
            p.ys[0].set(ysv)
            p.nx[0].set(nxv)
            p.ny[0].set(nyv)
            p.npair.set(1)

    def check(self, *args):
        """Callback function for running validity checks on the CCD
        parameters. It spots and flags overlapping windows, windows with null
        parameters, windows with invalid dimensions given the binning
        factors. It sets the correct number of windows according to the
        selected application and enables or disables the avalanche gain
        setting according to whether the avalanche output is being used.
        Finally it checks that the windows are synchronised and sets the
        status of the 'Sync' button accordingly.

        Returns True/False according to whether the settings are judged to be
        OK. True means they are thought to be in a fit state to be sent to the
        camera.

        This can only be run once the 'observe' are defined.
        """
        g = get_root(self).globals
        # Switch visible widget according to the application
        if self.isDrift():
            self.wframe.grid_forget()
            self.pframe.grid(row=2, column=0, columnspan=3, sticky=tk.W+tk.N)
            self.clearLab.config(state='disable')
            if not self.frozen:
                self.clear.config(state='disable')
                self.pframe.enable()
        else:
            self.pframe.grid_forget()
            self.wframe.grid(row=2, column=0, columnspan=3, sticky=tk.W+tk.N)
            self.clearLab.config(state='normal')
            if not self.frozen:
                self.clear.config(state='normal')
                self.wframe.enable()

        if self.avalanche():
            if not self.frozen:
                self.avgain.enable()
            if not self.oldAvalanche:
                # only update status if there has been a change
                # this is needed because any change to avGain causes
                # this check to be run and we must prevent the gain
                # automatically being set back to zero
                self.avgainLabel.configure(state='normal')
                self.avgain.set(0)
                self.oldAvalanche = True
        else:
            self.avgain.disable()
            self.avgainLabel.configure(state='disable')
            self.oldAvalanche = False

        # check the window settings
        if self.isDrift():
            status = self.pframe.check()
        else:
            status = self.wframe.check()

        # exposure delay
        if self.expose.ok():
            self.expose.config(bg=g.COL['main'])
        else:
            self.expose.config(bg=g.COL['warn'])
            status = False

        if status:
            # if valid, update timing and SN info
            g.count.update()

        return status

    def freeze(self):
        """
        Freeze all settings so that they can't be altered
        """
        self.app.disable()
        self.clear.disable()
        self.avalanche.disable()
        self.avgain.disable()
        self.readSpeed.disable()
        self.led.disable()
        self.expose.disable()
        self.number.disable()
        self.wframe.freeze()
        self.pframe.freeze()
        self.frozen = True

    def unfreeze(self):
        """
        Reverse of freeze
        """
        self.app.enable()
        self.clear.enable()
        self.avalanche.enable()
        self.readSpeed.enable()
        self.led.enable()
        self.expose.enable()
        self.number.enable()
        self.wframe.unfreeze()
        self.pframe.unfreeze()
        self.frozen = False
        self.check()

    def getRtplotWins(self):
        """
        Returns a string suitable to sending off to rtplot when
        it asks for window parameters. Returns null string '' if
        the windows are not OK. This operates on the basis of
        trying to send something back, even if it might not be
        OK as a window setup. Note that we have to take care
        here not to update any GUI components because this is
        called outside of the main thread.
        """
        try:
            xbin = self.wframe.xbin.value()
            ybin = self.wframe.ybin.value()
            if self.app.value() == 'Windows':
                nwin = self.wframe.nwin.value()
                ret = str(xbin) + ' ' + str(ybin) + ' ' + str(nwin) + '\r\n'
                for xs, ys, nx, ny in self.wframe:
                    ret += (str(xs) + ' ' + str(ys) + ' ' + str(nx) + ' ' +
                            str(ny) + '\r\n')
            elif self.app.value() == 'Drift':
                ret = str(xbin) + ' ' + str(ybin) + ' 2\r\n'
                for xsl, xsr, ys, nx, ny in self.pframe:
                    ret += (str(xsl) + ' ' + str(ys) + ' ' + str(nx) + ' ' +
                            str(ny) + '\r\n')
                    ret += (str(xsr) + ' ' + str(ys) + ' ' + str(nx) + ' ' +
                            str(ny) + '\r\n')

            return ret
        except Exception:
            return ''

    def timing(self):
        """
        Estimates timing information for the current setup. You should
        run a check on the instrument parameters before calling this.

        Returns: (expTime, deadTime, cycleTime, dutyCycle)

        expTime   : exposure time per frame (seconds)
        deadTime  : dead time per frame (seconds)
        cycleTime : sampling time (cadence), (seconds)
        dutyCycle : percentage time exposing.
        frameRate : number of frames per second
        """

        # avalanche mode y/n?
        lnormal = not self.avalanche()
        HCLOCK = HCLOCK_NORM if lnormal else HCLOCK_AV

        # drift mode y/n?
        isDriftMode = self.app.value() == 'Drift'

        # Set the readout speed
        readSpeed = self.readSpeed.value()

        if readSpeed == 'Fast':
            video = VIDEO_NORM_FAST if lnormal else VIDEO_AV_FAST
        elif readSpeed == 'Medium':
            video = VIDEO_NORM_MED if lnormal else VIDEO_AV_MED
        elif readSpeed == 'Slow':
            video = VIDEO_NORM_SLOW if lnormal else VIDEO_AV_SLOW
        else:
            raise ValueError('uspec.InstPars.timing: readout speed = '
                             + readSpeed + ' not recognised.')

        # clear chip on/off?
        lclear = not isDriftMode and self.clear()

        # get exposure delay
        expose = self.expose.value()

        # window parameters
        if isDriftMode:
            xbin = self.pframe.xbin.value()
            ybin = self.pframe.ybin.value()
            dxleft = self.pframe.xsl[0].value()
            dxright = self.pframe.xsr[0].value()
            dys = self.pframe.ys[0].value()
            dnx = self.pframe.nx[0].value()
            dny = self.pframe.ny[0].value()
        else:
            xbin = self.wframe.xbin.value()
            ybin = self.wframe.ybin.value()
            xs, ys, nx, ny = [], [], [], []
            nwin = self.wframe.nwin.value()
            for xsv, ysv, nxv, nyv in self.wframe:
                xs.append(xsv)
                ys.append(ysv)
                nx.append(nxv)
                ny.append(nyv)

        if lnormal:
            # normal mode convert xs by ignoring 16 overscan pixel
            if isDriftMode:
                dxleft += 16
                dxright += 16
            else:
                for nw in range(nwin):
                    xs[nw] += 16
        else:
            if isDriftMode:
                dxright = FFX - (dxright-1) - (dnx-1)
                dxleft = FFX - (dxleft-1) - (dnx-1)
                # in drift mode, also need to swap the windows around
                dxright, dxleft = dxleft, dxright
            else:
                # in avalanche mode, need to swap windows around
                for nw in range(nwin):
                    xs[nw] = FFX - (xs[nw]-1) - (nx[nw]-1)

        # convert timing parameters to seconds
        expose_delay = expose

        # clear chip by VCLOCK-ing the image and storage areas
        if lclear:
            # accomodate changes to clearing made by DA to fix dark current
            # when clearing charge along normal output
            clear_time = (2.0*(FFY*VCLOCK+39.e-6) + FFX*HCLOCK_NORM +
                          2162.0*HCLOCK_AV)
        else:
            clear_time = 0.0

        hclockFactor = 1.0 if lnormal else 2.0

        if isDriftMode:
            # for drift mode, we need the number of windows in the pipeline
            # and the pipeshift
            pnwin = int(((1037. / dny) + 1.)/2.)
            pshift = 1037. - (2.*pnwin-1.)*dny
            frame_transfer = (dny+dys-1.)*VCLOCK + 49.0e-6

            yshift = [0.]
            yshift[0] = (dys-1.0)*VCLOCK

            # After placing the window adjacent to the serial register, the
            # register must be cleared by clocking out the entire register,
            # taking FFX hclocks (we no longer open the dump gates, which
            # took only 8 hclock cycles to complete, but gave ramps and
            # bright rows in the bias). We think dave does 2*FFX hclocks
            # in avalanche mode, but need to check this with him.
            line_clear = [0.]
            if yshift[0] != 0:
                line_clear[0] = hclockFactor*FFX*HCLOCK

            numhclocks = [0]
            numhclocks[0] = FFX
            if not lnormal:
                numhclocks[0] += AVALANCHE_PIXELS

            line_read = [0.]
            line_read[0] = (VCLOCK*ybin + numhclocks[0]*HCLOCK +
                            video*2.0*dnx/xbin)

            readout = [0.]
            readout[0] = (dny/ybin) * line_read[0]

        else:
            # If not drift mode, move entire image into storage area
            # the -35 component is because Derek only shifts 1037 pixels
            # (composed of 1024 active rows, 5 dark reference rows, 2
            # transition rows and 6 extra overscan rows for good measure)
            # If drift mode, just move the window into the storage area
            frame_transfer = (FFY-35)*VCLOCK + 49.0e-6

            yshift = nwin*[0.]
            yshift[0] = (ys[0]-1.0)*VCLOCK
            for nw in range(1, nwin):
                yshift[nw] = (ys[nw]-ys[nw-1]-ny[nw-1])*VCLOCK

            line_clear = nwin*[0.]
            for nw in range(nwin):
                if yshift[nw] != 0:
                    line_clear[nw] = hclockFactor*FFX*HCLOCK

            # calculate how long it takes to shift one row into the serial
            # register shift along serial register and then read out the data.
            # The charge in a row after a window used to be dumped, taking
            # 8 HCLOCK cycles. This created ramps and bright rows/columns in
            # the images, so was removed.
            numhclocks = nwin*[0]
            for nw in range(nwin):
                numhclocks[nw] = FFX
                if not lnormal:
                    numhclocks[nw] += AVALANCHE_PIXELS

            line_read = nwin*[0.]
            for nw in range(nwin):
                line_read[nw] = (VCLOCK*ybin + numhclocks[nw]*HCLOCK +
                                 video*nx[nw]/xbin)

            # multiply time to shift one row into serial register by
            # number of rows for total readout time
            readout = nwin*[0.]
            for nw in range(nwin):
                readout[nw] = (ny[nw]/ybin) * line_read[nw]

        # now get the total time to read out one exposure.
        cycleTime = expose_delay + clear_time + frame_transfer
        if isDriftMode:
            cycleTime += pshift*VCLOCK+yshift[0]+line_clear[0]+readout[0]
        else:
            for nw in range(nwin):
                cycleTime += yshift[nw] + line_clear[nw] + readout[nw]

        frameRate = 1.0/cycleTime
        expTime = expose_delay if lclear else cycleTime - frame_transfer
        deadTime = cycleTime - expTime
        dutyCycle = 100.0*expTime/cycleTime

        return (expTime, deadTime, cycleTime, dutyCycle, frameRate)


class CountsFrame(tk.LabelFrame):
    """
    Frame for count rate estimates
    """
    def __init__(self, master):
        """
        master : enclosing widget
        """
        tk.LabelFrame.__init__(
            self, master, pady=2, text='Count & S-to-N estimator')

        # divide into left and right frames
        lframe = tk.Frame(self, padx=2)
        rframe = tk.Frame(self, padx=2)

        # entries
        self.filter = w.Radio(
            lframe, ('u', 'g', 'r', 'i', 'z'), 3, self.checkUpdate, initial=1)
        self.mag = w.RangedFloat(
            lframe, 18., 0., 30., self.checkUpdate, True, width=5)
        self.seeing = w.RangedFloat(
            lframe, 1.0, 0.2, 20., self.checkUpdate, True, True, width=5)
        self.airmass = w.RangedFloat(
            lframe, 1.5, 1.0, 5.0, self.checkUpdate, True, width=5)
        self.moon = w.Radio(lframe, ('d', 'g', 'b'),  3, self.checkUpdate)

        # results
        self.cadence = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.exposure = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.duty = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.peak = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.total = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.ston = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)
        self.ston3 = w.Ilabel(rframe, text='UNDEF', width=10, anchor=tk.W)

        # layout
        # left
        tk.Label(lframe, text='Filter:').grid(
            row=0, column=0, padx=5, pady=3, sticky=tk.W+tk.N)
        self.filter.grid(row=0, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text='Mag:').grid(
            row=1, column=0, padx=5, pady=3, sticky=tk.W)
        self.mag.grid(row=1, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text='Seeing:').grid(
            row=2, column=0, padx=5, pady=3, sticky=tk.W)
        self.seeing.grid(row=2, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text='Airmass:').grid(
            row=3, column=0, padx=5, pady=3, sticky=tk.W)
        self.airmass.grid(row=3, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text='Moon:').grid(
            row=4, column=0, padx=5, pady=3, sticky=tk.W)
        self.moon.grid(row=4, column=1, padx=5, pady=3, sticky=tk.W)

        # right
        tk.Label(rframe, text='Cadence:').grid(
            row=0, column=0, padx=5, pady=3, sticky=tk.W)
        self.cadence.grid(row=0, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='Exposure:').grid(
            row=1, column=0, padx=5, pady=3, sticky=tk.W)
        self.exposure.grid(row=1, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='Duty cycle:').grid(
            row=2, column=0, padx=5, pady=3, sticky=tk.W)
        self.duty.grid(row=2, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='Peak:').grid(
            row=3, column=0, padx=5, pady=3, sticky=tk.W)
        self.peak.grid(row=3, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='Total:').grid(
            row=4, column=0, padx=5, pady=3, sticky=tk.W)
        self.total.grid(row=4, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='S/N:').grid(
            row=5, column=0, padx=5, pady=3, sticky=tk.W)
        self.ston.grid(row=5, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text='S/N (3h):').grid(
            row=6, column=0, padx=5, pady=3, sticky=tk.W)
        self.ston3.grid(row=6, column=1, padx=5, pady=3, sticky=tk.W)

        # slot frames in
        lframe.grid(row=0, column=0, sticky=tk.W+tk.N)
        rframe.grid(row=0, column=1, sticky=tk.W+tk.N)

    def checkUpdate(self, *args):
        """
        Updates values after first checking instrument parameters are OK.
        This is not integrated within update to prevent ifinite recursion
        since update gets called from ipars.
        """
        g = get_root(self).globals
        if not self.check():
            g.clog.warn('Current observing parameters are not valid.')
            return False

        if not g.ipars.check():
            g.clog.warn('Current instrument parameters are not valid.')
            return False

    def check(self):
        """
        Checks values
        """
        status = True
        g = get_root(self).globals
        if self.mag.ok():
            self.mag.config(bg=g.COL['main'])
        else:
            self.mag.config(bg=g.COL['warn'])
            status = False

        if self.airmass.ok():
            self.airmass.config(bg=g.COL['main'])
        else:
            self.airmass.config(bg=g.COL['warn'])
            status = False

        if self.seeing.ok():
            self.seeing.config(bg=g.COL['main'])
        else:
            self.seeing.config(bg=g.COL['warn'])
            status = False

        return status

    def update(self, *args):
        """
        Updates values. You should run a check on the instrument and
        target parameters before calling this.
        """
        g = get_root(self).globals
        expTime, deadTime, cycleTime, dutyCycle, frameRate = g.ipars.timing()

        total, peak, peakSat, peakWarn, ston, ston3 = \
            self.counts(expTime, cycleTime)

        if cycleTime < 0.01:
            self.cadence.config(text='{0:7.5f} s'.format(cycleTime))
        elif cycleTime < 0.1:
            self.cadence.config(text='{0:6.4f} s'.format(cycleTime))
        elif cycleTime < 1.:
            self.cadence.config(text='{0:5.3f} s'.format(cycleTime))
        elif cycleTime < 10.:
            self.cadence.config(text='{0:4.2f} s'.format(cycleTime))
        elif cycleTime < 100.:
            self.cadence.config(text='{0:4.1f} s'.format(cycleTime))
        elif cycleTime < 1000.:
            self.cadence.config(text='{0:4.0f} s'.format(cycleTime))
        else:
            self.cadence.config(text='{0:5.0f} s'.format(cycleTime))

        if expTime < 0.01:
            self.exposure.config(text='{0:7.5f} s'.format(expTime))
        elif expTime < 0.1:
            self.exposure.config(text='{0:6.4f} s'.format(expTime))
        elif expTime < 1.:
            self.exposure.config(text='{0:5.3f} s'.format(expTime))
        elif expTime < 10.:
            self.exposure.config(text='{0:4.2f} s'.format(expTime))
        elif expTime < 100.:
            self.exposure.config(text='{0:4.1f} s'.format(expTime))
        elif expTime < 1000.:
            self.exposure.config(text='{0:4.0f} s'.format(expTime))
        else:
            self.exposure.config(text='{0:5.0f} s'.format(expTime))

        self.duty.config(text='{0:4.1f} %'.format(dutyCycle))
        self.peak.config(text='{0:d} cts'.format(int(round(peak))))
        if peakSat:
            self.peak.config(bg=g.COL['error'])
        elif peakWarn:
            self.peak.config(bg=g.COL['warn'])
        else:
            self.peak.config(bg=g.COL['main'])

        self.total.config(text='{0:d} cts'.format(int(round(total))))
        self.ston.config(text='{0:.1f}'.format(ston))
        self.ston3.config(text='{0:.1f}'.format(ston3))

    def counts(self, expTime, cycleTime, ap_scale=1.6, ndiv=5):
        """
        Computes counts per pixel, total counts, sky counts
        etc given current magnitude, seeing etc. You should
        run a check on the instrument parameters before calling
        this.

        expTime   : exposure time per frame (seconds)
        cycleTime : sampling, cadence (seconds)
        ap_scale  : aperture radius as multiple of seeing

        Returns: (total, peak, peakSat, peakWarn, ston, ston3)

        total    -- total number of object counts in aperture
        peak     -- peak counts in a pixel
        peakSat  -- flag to indicate saturation
        peakWarn -- flag to indication level approaching saturation
        ston     -- signal-to-noise per exposure
        ston3    -- signal-to-noise after 3 hours on target
        """

        # code directly translated from Java equivalent.
        g = get_root(self).globals
        # avalanche mode y/n?
        lnormal = not g.ipars.avalanche()

        # Set the readout speed
        readSpeed = g.ipars.readSpeed.value()

        if g.ipars.app == 'Windows':
            xbin, ybin = g.ipars.wframe.xbin.value(), g.ipars.wframe.ybin.value()
        else:
            xbin, ybin = g.ipars.pframe.xbin.value(), g.ipars.pframe.ybin.value()

        # calculate SN info.
        zero, sky, skyTot, gain, read, darkTot = 0., 0., 0., 0., 0., 0.
        total, peak, correct, signal, readTot, seeing = 0., 0., 0., 0., 0., 0.
        noise, _, narcsec, npix, signalToNoise3 = 1., 0., 0., 0., 0.

        tinfo = g.TINS[g.cpars['telins_name']]
        filtnam = self.filter.value()
        zero = tinfo['zerop'][filtnam]
        mag = self.mag.value()
        seeing = self.seeing.value()
        sky = g.SKY[self.moon.value()][filtnam]
        airmass = self.airmass.value()

        # GAIN, RNO
        if readSpeed == 'Fast':
            gain = GAIN_NORM_FAST if lnormal else GAIN_AV_FAST
            read = RNO_NORM_FAST if lnormal else RNO_AV_FAST

        elif readSpeed == 'Medium':
            gain = GAIN_NORM_MED if lnormal else GAIN_AV_MED
            read = RNO_NORM_MED if lnormal else RNO_AV_MED

        elif readSpeed == 'Slow':
            gain = GAIN_NORM_SLOW if lnormal else GAIN_AV_SLOW
            read = RNO_NORM_SLOW if lnormal else RNO_AV_SLOW

        plateScale = tinfo['plateScale']

        # calculate expected electrons
        total = 10.**((zero-mag-airmass*g.EXTINCTION[filtnam])/2.5)*expTime

        # compute fraction that fall in central pixel
        # assuming target exactly at its centre. Do this
        # by splitting each pixel of the central (potentially
        # binned) pixel into ndiv * ndiv points at
        # which the seeing profile is added. sigma is the
        # RMS seeing in terms of pixels.
        sigma = seeing/g.EFAC/plateScale

        sum = 0.
        for iyp in range(ybin):
            yoff = -ybin/2.+iyp
            for ixp in range(xbin):
                xoff = -xbin/2.+ixp
                for iys in range(ndiv):
                    y = (yoff + (iys+0.5)/ndiv)/sigma
                    for ixs in range(ndiv):
                        x = (xoff + (ixs+0.5)/ndiv)/sigma
                        sum += math.exp(-(x*x+y*y)/2.)
        peak = total*sum/(2.*math.pi*sigma**2*ndiv**2)
        # peak    = total*xbin*ybin*(plateScale/(seeing/EFAC))**2/(2.*math.pi)

        # Work out fraction of flux in aperture with radius AP_SCALE*seeing
        correct = 1. - math.exp(-(g.EFAC*ap_scale)**2/2.)

        # expected sky e- per arcsec
        skyPerArcsec = 10.**((zero-sky)/2.5)*expTime
        # skyPerPixel = skyPerArcsec*plateScale**2*xbin*ybin
        narcsec = math.pi*(ap_scale*seeing)**2
        skyTot = skyPerArcsec*narcsec
        npix = math.pi*(ap_scale*seeing/plateScale)**2/xbin/ybin

        signal = correct*total        # in electrons
        darkTot = npix*DARK_E*expTime  # in electrons
        readTot = npix*read**2         # in electrons
        cic = 0 if lnormal else CIC

        # noise, in electrons
        if lnormal:
            noise = math.sqrt(readTot + darkTot + skyTot + signal + cic)
        else:
            # assume high gain observations in proportional mode
            noise = math.sqrt(readTot/AVALANCHE_GAIN_9**2 +
                              2.0*(darkTot + skyTot + signal) + cic)

        # Now compute signal-to-noise in 3 hour seconds run
        signalToNoise3 = signal/noise*math.sqrt(3*3600./cycleTime)

        # if using the avalanche mode, check that the signal level
        # is safe. A single electron entering the avalanche register
        # results in a distribution of electrons at the output with
        # mean value given by the parameter avalanche_gain. The
        # distribution is close to exponential, hence the probability
        # of obtaining an amplification n times higher than the mean is
        # given by e**-n. A value of 3/5 for n is adopted here for
        # warning/safety, which will occur once in every ~20/100
        # amplifications

        # convert from electrons to counts
        total /= gain
        peak /= gain

        warn = 25000
        sat = 60000

        if not lnormal:
            sat = AVALANCHE_SATURATE/AVALANCHE_GAIN_9/5/gain
            warn = AVALANCHE_SATURATE/AVALANCHE_GAIN_9/3/gain

        peakSat = peak > sat
        peakWarn = peak > warn

        return (total, peak, peakSat, peakWarn, signal/noise, signalToNoise3)



# Specific widgets, widget groups and parameters for hipercam instrument
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import math
import pickle
from os.path import expanduser

# non-standard imports
import numpy as np
import six
from twisted.internet.defer import inlineCallbacks, returnValue

# internal imports
from . import DriverError
from . import widgets as w
from .misc import (
    ReadNGCTelemetry,
    createJSON,
    execCommand,
    isPoweredOn,
    isRunActive,
    jsonFromFits,
    postJSON,
    saveJSON,
    startNodding,
)
from .tkutils import get_root, place_at_edge

if not six.PY3:
    import tkFileDialog as filedialog
    import Tkinter as tk
    import tkMessageBox as messagebox
else:
    import tkinter as tk
    from tkinter import filedialog, messagebox


# Timing, Gain, Noise parameters
# Times in seconds
VCLOCK_FRAME_SLOW = 15e-6  # vertical clocking time in image area
VCLOCK_STORAGE_SLOW = 20e-6  # vertical clocking time in storage area
HCLOCK_SLOW = 0.24e-6  # horizontal clocking time
VCLOCK_FAST = 13e-6  # faster mode with poorer CTE
HCLOCK_FAST = 0.12e-6  # faster mode with poorer CTE
SETUP_READ = 9.0e-7  # time required for Naidu's setup_read SR
DUMP_TIME_SLOW = 18e-6  # time to dump extra pixels, slow clocking
DUMP_TIME_FAST = 3.6e-6  # time to dump extra pixels, fast clocking
VIDEO_SLOW_SE = 8.72e-6  # ~113 kHz, Naidu's clock speed for single output mode
VIDEO_SLOW = 5.2e-6  # ~192 kHz, same clock speed as fast, but 4 samples
VIDEO_FAST = 1.9e-6  # ~520 kHz
GAIN_FAST = 1.1  # electrons/ADU
GAIN_SLOW = 1.1
RNO_FAST = 5.0  # e- / pixel
RNO_SLOW = 4.5
DARK_E = 0.02  # e/pix/s


FFX = 1024  # X pixels per output
FFY = 520  # Y pixels per output
PRSCX = 50  # number of pre-scan pixels


class ExposureMultiplier(tk.LabelFrame):
    """
    Top to bottom group of RangedInt entry items to specify Nblue etc. Has a max
    number of rows after which it will jump to the left of next column and start over.
    """

    def __init__(
        self, master, labels, ivals, imins, imaxs, nrmax, checker, blank, **kw
    ):
        """
        Parameters
        ----------
        master : tk.widget
            enclosing widget
        labels : iterable
            labels for entry items
        ivals : list of int
            initial values
        imins : list of int
            minimum values
        imaxs : list of int
            maximum values
        nrmax : int
            maximum number of rows before wrapping
        checker : callable
            command that is run on any change to the entry
        blank : bool
            controls whether the field is allowed to be blank
            In some cases it makes things easier if blank fields are
            allowed, even if it is technically invalid
        kw : dict
            keyword arguments

        """
        tk.LabelFrame.__init__(self, master, bd=0)
        nitems = len(labels)
        if len(ivals) != nitems or len(imins) != nitems or len(imaxs) != nitems:
            raise DriverError(
                "ExposureMultiplier.__init__ values and options "
                + "must have same length"
            )
        self.nitems = nitems
        self.labels = labels
        self.widgets = [
            w.RangedInt(self, ival, imin, imax, checker, blank, **kw)
            for ival, imin, imax in zip(ivals, imins, imaxs)
        ]
        row = 0
        col = 0
        for nw, widget in enumerate(self.widgets):
            tk.Label(self, text=labels[nw]).grid(row=row, column=col, sticky=tk.W)
            widget.grid(row=row, column=col + 1, sticky=tk.W)
            row += 1
            if row == nrmax:
                col += 1
                row = 0

    def value(self, index):
        return self.widgets[index].value()

    def set(self, index, num):
        self.widgets[index].set(num)

    def getall(self):
        return [widget.value() for widget in self.widgets]

    def setall(self, nums):
        for index, val in enumerate(nums):
            self.set(index, val)

    def disable(self):
        for widget in self.widgets:
            widget.configure(state="disable")
            widget.set_unbind()

    def enable(self):
        for widget in self.widgets:
            widget.configure(state="normal")
            widget.set_bind()


class InstPars(tk.LabelFrame):
    """
    Instrument parameters block.

    This widget block contains the meat of hdriver. Window settings, readout speed
    etc.
    """

    def __init__(self, master):
        """
        master : enclosing widget
        """
        tk.LabelFrame.__init__(self, master)

        # left hand side
        lhs = tk.Frame(self)
        rhs = tk.Frame(self)

        # Application (mode)
        tk.Label(lhs, text="Mode").grid(row=0, column=0, sticky=tk.W)
        self.app = w.Radio(
            lhs,
            ("Full", "Wins", "Drift"),
            3,
            self.check,
            ("FullFrame", "Windows", "Drift"),
        )
        self.app.grid(row=0, column=1, columnspan=2, sticky=tk.W)

        # Clear enabled
        self.clearLab = tk.Label(lhs, text="Clear")
        self.clearLab.grid(row=1, column=0, sticky=tk.W)
        self.clear = w.OnOff(lhs, False, self.check)
        self.clear.grid(row=1, column=1, columnspan=2, sticky=tk.W)

        # nod telescope
        self.nodLab = tk.Label(lhs, text="Dithering")
        self.nodLab.grid(row=2, column=0, sticky=tk.W)
        self.nod = w.OnOff(lhs, False, self.setupNodding)
        self.nod.grid(row=2, column=1, columnspan=2, sticky=tk.W)
        self.nodPattern = {}

        # Overscan in x enabled
        self.oscanLab = tk.Label(lhs, text="Overscan")
        self.oscanLab.grid(row=3, column=0, sticky=tk.W)
        self.oscan = w.OnOff(lhs, False, self.check)
        self.oscany = w.OnOff(lhs, False, self.check)
        self.oscan.grid(row=3, column=1, sticky=tk.W)
        self.oscany.grid(row=3, column=2, sticky=tk.W)

        # led on (expert mode only)
        self.ledLab = tk.Label(lhs, text="LED setting")
        self.ledLab.grid(row=4, column=0, sticky=tk.W)
        self.led = w.OnOff(lhs, False, None)
        self.led.grid(row=4, column=1, columnspan=2, pady=2, sticky=tk.W)

        # dummy mode enabled (expert mode only)
        self.dummyLab = tk.Label(lhs, text="Dummy Output")
        self.dummyLab.grid(row=5, column=0, sticky=tk.W)
        self.dummy = w.OnOff(lhs, True, self.check)
        self.dummy.grid(row=5, column=1, columnspan=2, pady=2, sticky=tk.W)

        # Faster Clock speed enabled
        self.fastClkLab = tk.Label(lhs, text="Fast Clocks")
        self.fastClkLab.grid(row=6, column=0, sticky=tk.W)
        self.fastClk = w.OnOff(lhs, False, self.check)
        self.fastClk.grid(row=6, column=1, columnspan=2, pady=2, sticky=tk.W)

        # Readout speed
        tk.Label(lhs, text="Readout speed").grid(row=7, column=0, sticky=tk.W)
        self.readSpeed = w.Select(lhs, 1, ("Fast", "Slow"), self.check)
        self.readSpeed.grid(row=7, column=1, columnspan=2, pady=2, sticky=tk.W)

        # Exp delay
        tk.Label(lhs, text="Exposure delay (s)").grid(row=8, column=0, sticky=tk.W)
        self.expose = w.RangedFloat(
            lhs,
            0.1,
            0.00001,
            1677.7207,
            self.check,
            blank=True,
            allowzero=True,
            width=7,
        )
        self.expose.grid(row=8, column=1, columnspan=2, pady=2, sticky=tk.W)

        # num exp
        tk.Label(lhs, text="Num. exposures  ").grid(row=9, column=0, sticky=tk.W)
        self.number = w.PosInt(lhs, 0, None, False, width=7)
        self.number.grid(row=9, column=1, columnspan=2, pady=2, sticky=tk.W)

        # nb, ng, nr etc
        labels = ("nu", "ng", "nr", "ni", "nz")
        ivals = (1, 1, 1, 1, 1)
        imins = (1, 1, 1, 1, 1)
        imaxs = (500, 500, 500, 500, 500)
        self.nmult = ExposureMultiplier(
            rhs, labels, ivals, imins, imaxs, 5, self.check, False, width=4
        )
        # grid (on RHS)
        self.nmult.grid(row=0, column=0, columnspan=2, pady=2, sticky=tk.E + tk.S)

        tk.Label(rhs, text="COMPO  ").grid(row=1, column=0)
        self.compo = w.OnOff(rhs, False, self.check)
        self.compo.grid(row=1, column=1, pady=2, sticky=tk.W)

        # We have two possible window frames. A single pair for
        # drift mode, or a 2-quad frame for window mode.

        # drift frame
        # xstart - LH window of pair
        xsls = (1,)
        xslmins = (1,)
        xslmaxs = (1024,)
        # xstart - RH window of pair
        xsrs = (1025,)
        xsrmins = (1025,)
        xsrmaxs = (2048,)
        # ystart values
        yss = (1,)
        ysmins = (1,)
        ysmaxs = (512,)
        # sizes of windows (start at FF)
        nxs = (100,)
        nys = (100,)
        # allowed binning factors
        xbfac = tuple(range(1, 21))
        ybfac = tuple(range(1, 21))
        self.drift_frame = w.WinPairs(
            lhs,
            xsls,
            xslmins,
            xslmaxs,
            xsrs,
            xsrmins,
            xsrmaxs,
            yss,
            ysmins,
            ysmaxs,
            nxs,
            nys,
            xbfac,
            ybfac,
            self.check,
        )

        # window frame for quads
        # xstart on LHS
        xsll = xsul = (1, 1)
        xsllmin = xsulmin = (1, 1)
        xsllmax = xsulmax = (1024, 1024)
        # xstart on RHS
        xslr = xsur = (1025, 1025)
        xslrmin = xsurmin = (1025, 1025)
        xslrmax = xsurmax = (2048, 2048)
        # ystart
        ys = (1, 1)
        ysmin = (1, 1)
        ysmax = (512, 512)
        # sizes (start at FF)
        nx = (1024, 1024)
        ny = (512, 512)
        self.quad_frame = w.WinQuads(
            lhs,
            xsll,
            xsllmin,
            xsllmax,
            xsul,
            xsulmin,
            xsulmax,
            xslr,
            xslrmin,
            xslrmax,
            xsur,
            xsurmin,
            xsurmax,
            ys,
            ysmin,
            ysmax,
            nx,
            ny,
            xbfac,
            ybfac,
            self.check,
        )

        self.quad_frame.grid(row=10, column=0, columnspan=3, sticky=tk.W + tk.N)

        # Pack two halfs
        lhs.pack(side=tk.LEFT, anchor=tk.N, padx=5)
        rhs.pack(side=tk.LEFT, anchor=tk.N, padx=5)

        # Store freeze state
        self.frozen = False

        self.setExpertLevel()

    @property
    def wframe(self):
        if self.isDrift():
            return self.drift_frame
        return self.quad_frame

    def setupNodding(self):
        """
        Setup Nodding for GTC
        """
        g = get_root(self).globals

        if not self.nod():
            # re-enable clear mode box if not drift
            if not self.isDrift():
                self.clear.enable()

            # clear existing nod pattern
            self.nodPattern = {}
            self.check()
            return

        # Do nothing if we're not at the GTC
        if g.cpars["telins_name"] != "GTC":
            messagebox.showerror("Error", "Cannot dither WHT")
            self.nod.set(False)
            self.nodPattern = {}
            return

        # check for drift mode and bomb out
        if self.isDrift():
            messagebox.showerror("Error", "Cannot dither telescope in drift mode")
            self.nod.set(False)
            self.nodPattern = {}
            return

        # check for clear not enabled and warn
        if not self.clear():
            if not messagebox.askokcancel(
                "Warning", "Dithering telescope will enable clear mode. Continue?"
            ):
                self.nod.set(False)
                self.nodPattern = {}
                return

        # Ask for nod pattern
        try:
            home = expanduser("~")
            fname = filedialog.askopenfilename(
                title="Open offsets text file",
                defaultextension=".txt",
                filetypes=[("text files", ".txt")],
                initialdir=home,
            )

            if not fname:
                g.clog.warn("Aborted load from disk")
                raise ValueError

            ra, dec = np.loadtxt(fname).T
            if len(ra) != len(dec):
                g.clog.warn("Mismatched lengths of RA and Dec offsets")
                raise ValueError

            data = dict(ra=ra.tolist(), dec=dec.tolist())
        except Exception:
            g.clog.warn("Setting dither pattern failed. Disabling dithering")
            self.nod.set(False)
            self.nodPattern = {}
            return

        # store nodding on ipars object
        self.nodPattern = data
        # enable clear mode
        self.clear.set(True)
        # update
        self.check()

    def setExpertLevel(self):
        g = get_root(self).globals
        level = g.cpars["expert_level"]
        if level == 0:
            self.ledLab.grid_forget()
            self.led.grid_forget()
            self.led.set(0)

            self.oscanLab.config(text="Overscan")
            self.oscany.grid_forget()
            self.remember_oscany = self.oscany()
            self.oscany.set(0)

            self.dummyLab.grid_forget()
            self.dummy.grid_forget()
        else:
            self.ledLab.grid(row=4, column=0, sticky=tk.W)
            self.led.grid(row=4, column=1, columnspan=2, pady=2, sticky=tk.W)

            self.oscanLab.config(text="Overscan (x, y)")
            self.oscany.grid(row=3, column=2, sticky=tk.W)
            if hasattr(self, "remember_oscany"):
                self.oscany.set(self.remember_oscany)

            self.dummyLab.grid(row=5, column=0, sticky=tk.W)
            self.dummy.grid(row=5, column=1, columnspan=2, pady=2, sticky=tk.W)

    def isDrift(self):
        if self.app.value() == "Drift":
            return True
        elif self.app.value() in ["FullFrame", "Windows"]:
            return False
        else:
            raise DriverError(
                "InstPars.isDrift: application " + self.app.value() + " not recognised"
            )

    def isFF(self):
        if self.app.value() == "FullFrame":
            return True
        elif self.app.value() in ["Drift", "Windows"]:
            return False
        else:
            raise DriverError(
                "InstPars.isDrift: application " + self.app.value() + " not recognised"
            )

    def dumpJSON(self):
        """
        Encodes current parameters to JSON compatible dictionary
        """
        numexp = self.number.get()
        expTime, _, cadence, _, _ = self.timing()
        if numexp == 0:
            numexp = -1

        data = dict(
            numexp=self.number.value(),
            app=self.app.value(),
            led_flsh=self.led(),
            dummy_out=self.dummy(),
            fast_clks=self.fastClk(),
            readout=self.readSpeed(),
            dwell=self.expose.value(),
            exptime=expTime,
            cadence=cadence,
            oscan=self.oscan(),
            oscany=self.oscany(),
            xbin=self.wframe.xbin.value(),
            ybin=self.wframe.ybin.value(),
            multipliers=self.nmult.getall(),
            clear=self.clear(),
        )

        # only allow nodding in clear mode, even if GUI has got confused
        if data["clear"] and self.nodPattern:
            data["nodpattern"] = self.nodPattern

        # no mixing clear and multipliers, no matter what GUI says
        if data["clear"]:
            data["multipliers"] = [1 for i in self.nmult.getall()]

        # add window mode
        if not self.isFF():
            if self.isDrift():
                # no clear, multipliers or oscan in drift
                for setting in ("clear", "oscan", "oscany"):
                    data[setting] = 0
                data["multipliers"] = [1 for i in self.nmult.getall()]

                for iw, (xsl, xsr, ys, nx, ny) in enumerate(self.wframe):
                    data["x{}start_left".format(iw + 1)] = xsl
                    data["x{}start_right".format(iw + 1)] = xsr
                    data["y{}start".format(iw + 1)] = ys
                    data["y{}size".format(iw + 1)] = ny
                    data["x{}size".format(iw + 1)] = nx
            else:
                # no oscany in window mode
                data["oscany"] = 0

                for iw, (xsll, xsul, xslr, xsur, ys, nx, ny) in enumerate(self.wframe):
                    data["x{}start_upperleft".format(iw + 1)] = xsul
                    data["x{}start_lowerleft".format(iw + 1)] = xsll
                    data["x{}start_upperright".format(iw + 1)] = xsur
                    data["x{}start_lowerright".format(iw + 1)] = xslr
                    data["y{}start".format(iw + 1)] = ys
                    data["x{}size".format(iw + 1)] = nx
                    data["y{}size".format(iw + 1)] = ny
        return data

    def loadJSON(self, json_string):
        """
        Loads in an application saved in JSON format.
        """
        g = get_root(self).globals

        # enable COMPO if present in JSON
        if "compo" in json.loads(json_string):
            self.compo.set(1)
        else:
            self.compo.set(0)

        data = json.loads(json_string)["appdata"]
        # first set the parameters which change regardless of mode
        # number of exposures
        numexp = data.get("numexp", 0)
        if numexp == -1:
            numexp = 0
        self.number.set(numexp)
        # Overscan (x, y)
        if "oscan" in data:
            self.oscan.set(data["oscan"])
        if "oscany" in data:
            self.oscan.set(data["oscany"])
        # LED setting
        self.led.set(data.get("led_flsh", 0))
        # Dummy output enabled
        self.dummy.set(data.get("dummy_out", 1))
        # Fast clocking option?
        self.fastClk.set(data.get("fast_clks", 0))
        # readout speed
        self.readSpeed.set(data.get("readout", "Slow"))
        # dwell
        dwell = data.get("dwell", 0)
        self.expose.set(str(float(dwell)))

        # multipliers
        mult_values = data.get("multipliers", (1, 1, 1, 1, 1))
        self.nmult.setall(mult_values)

        # look for nodpattern in data
        nodPattern = data.get("nodpattern", {})
        if nodPattern and g.cpars["telins_name"] == "GTC":
            self.nodPattern = nodPattern
            self.nod.set(True)
            self.clear.set(True)
        else:
            self.nodPattern = {}
            self.nod.set(False)

        # binning
        self.quad_frame.xbin.set(data.get("xbin", 1))
        self.quad_frame.ybin.set(data.get("ybin", 1))
        self.drift_frame.xbin.set(data.get("xbin", 1))
        self.drift_frame.ybin.set(data.get("ybin", 1))

        # now for the behaviour which depends on mode
        if "app" in data:
            self.app.set(data["app"])
            app = data["app"]

            if app == "Drift":
                # disable clear mode in drift
                self.clear.set(0)
                # only one pair allowed
                self.wframe.npair.set(1)

                # set the window pair values
                labels = (
                    "x1start_left",
                    "y1start",
                    "x1start_right",
                    "x1size",
                    "y1size",
                )
                if not all(label in data for label in labels):
                    raise DriverError("Drift mode application missing window params")
                # now actually set them
                self.wframe.xsl[0].set(data["x1start_left"])
                self.wframe.xsr[0].set(data["x1start_right"])
                self.wframe.ys[0].set(data["y1start"])
                self.wframe.nx[0].set(data["x1size"])
                self.wframe.ny[0].set(data["y1size"])
                self.wframe.check()

            elif app == "FullFrame":
                # enable clear mode if set
                self.clear.set(data.get("clear", 0))

            elif app == "Windows":
                # enable clear mode if set
                self.clear.set(data.get("clear", 0))
                nquad = 0
                for nw in range(2):
                    labels = (
                        (
                            "x{0}start_lowerleft y{0}start x{0}start_upperleft x{0}start_upperright "
                            + "x{0}start_lowerright x{0}size y{0}size"
                        )
                        .format(nw + 1)
                        .split()
                    )
                    if all(label in data for label in labels):
                        xsll = data[labels[0]]
                        xslr = data[labels[4]]
                        xsul = data[labels[2]]
                        xsur = data[labels[3]]
                        ys = data[labels[1]]
                        nx = data[labels[5]]
                        ny = data[labels[6]]
                        self.wframe.xsll[nw].set(xsll)
                        self.wframe.xslr[nw].set(xslr)
                        self.wframe.xsul[nw].set(xsul)
                        self.wframe.xsur[nw].set(xsur)
                        self.wframe.ys[nw].set(ys)
                        self.wframe.nx[nw].set(nx)
                        self.wframe.ny[nw].set(ny)
                        nquad += 1
                    else:
                        break
                self.wframe.nquad.set(nquad)
                self.wframe.check()

    @inlineCallbacks
    def check(self, *args):
        """
        Callback to check validity of instrument parameters.

        Performs the following tasks:
            - spots and flags overlapping windows or null window parameters
            - flags windows with invalid dimensions given the binning parameter
            - sets the correct number of enabled windows
            - disables or enables clear and nod buttons depending on drift mode or not
            - checks for window synchronisation, enabling sync button if required
            - enables or disables start button if settings are OK

        Returns
        -------
        status : bool
            True or False according to whether the settings are OK.
        """
        status = True
        root = get_root(self)
        g = root.globals

        # if we've just enabled COMPO, then raise window if exists
        if self.compo():
            compo_hw_widget = getattr(g, "compo_hw", None)
            if compo_hw_widget is not None:
                if compo_hw_widget.state() == "withdrawn":
                    compo_hw_widget.deiconify()
                    place_at_edge(root, compo_hw_widget)

        # clear errors on binning (may be set later if FF)
        xbinw, ybinw = self.wframe.xbin, self.wframe.ybin
        xbinw.config(bg=g.COL["main"])
        ybinw.config(bg=g.COL["main"])

        # keep binning factors of drift mode and windowed mode up to date
        oframe, aframe = (
            (self.quad_frame, self.drift_frame)
            if self.drift_frame.winfo_ismapped()
            else (self.drift_frame, self.quad_frame)
        )
        xbin, ybin = aframe.xbin.value(), aframe.ybin.value()
        oframe.xbin.set(xbin)
        oframe.ybin.set(ybin)

        if not self.frozen:
            if self.clear() or self.isDrift():
                # disable nmult in clear or drift mode
                self.nmult.disable()
            else:
                self.nmult.enable()

        if self.isDrift():
            self.clearLab.config(state="disable")
            self.nodLab.config(state="disable")
            if not self.drift_frame.winfo_ismapped():
                self.quad_frame.grid_forget()
                self.drift_frame.grid(
                    row=10, column=0, columnspan=3, sticky=tk.W + tk.N
                )

            if not self.frozen:
                self.oscany.config(state="disable")
                self.oscan.config(state="disable")
                self.clear.config(state="disable")
                self.nod.config(state="disable")
                self.wframe.enable()
                status = self.wframe.check()

        elif self.isFF():
            # special case check of binning from window frame
            if 1024 % xbin != 0:
                status = False
                xbinw.config(bg=g.COL["error"])
            elif (1024 // xbin) % 4 != 0:
                status = False
                xbinw.config(bg=g.COL["error"])
            if 512 % ybin != 0:
                status = False
                ybinw.config(bg=g.COL["error"])

            if not self.quad_frame.winfo_ismapped():
                self.drift_frame.grid_forget()
                self.quad_frame.grid(row=10, column=0, columnspan=3, sticky=tk.W + tk.N)

            self.clearLab.config(state="normal")
            if g.cpars["telins_name"] == "GTC":
                self.nodLab.config(state="normal")
            else:
                self.nodLab.config(state="disable")
            if not self.frozen:
                self.oscany.config(state="normal")
                self.oscan.config(state="normal")
                self.clear.config(state="normal")
                if g.cpars["telins_name"] == "GTC":
                    self.nod.config(state="normal")
                else:
                    self.nod.config(state="disable")
                self.wframe.disable()

        else:
            self.clearLab.config(state="normal")
            if g.cpars["telins_name"] == "GTC":
                self.nodLab.config(state="normal")
            else:
                self.nodLab.config(state="disable")
            if not self.quad_frame.winfo_ismapped():
                self.drift_frame.grid_forget()
                self.quad_frame.grid(row=10, column=0, columnspan=3, sticky=tk.W + tk.N)

            if not self.frozen:
                self.oscany.config(state="disable")
                self.oscan.config(state="normal")
                self.clear.config(state="normal")
                if g.cpars["telins_name"] == "GTC":
                    self.nod.config(state="normal")
                else:
                    self.nod.config(state="disable")
                self.wframe.enable()
                status = self.wframe.check()

        # exposure delay
        if self.expose.ok():
            self.expose.config(bg=g.COL["main"])
        else:
            self.expose.config(bg=g.COL["warn"])
            status = False

        # don't allow binning other than 1, 2 in overscan or prescan mode
        if self.oscan() or self.oscany():
            if xbin not in (1, 2):
                status = False
                xbinw.config(bg=g.COL["error"])
            if ybin not in (1, 2):
                status = False
                ybinw.config(bg=g.COL["error"])

        # disable clear if nodding enabled. re-enable if not drift
        if not self.frozen:
            if self.nod() or self.nodPattern:
                self.clear.config(state="disabled")
                self.clearLab.config(state="disabled")
            elif not self.isDrift():
                self.clear.config(state="normal")
                self.clearLab.config(state="normal")

        # allow posting if parameters are OK. update count and SN estimates too
        if status:
            try:
                run_active = yield isRunActive(g)
                powered_on = yield isPoweredOn(g)
            except Exception as err:
                g.clog.warn(str(err))
            if (
                g.cpars["hcam_server_on"]
                and g.cpars["eso_server_online"]
                and g.observe.start["state"] == "disabled"
                and not run_active
                and powered_on
            ):
                g.observe.start.enable()
            g.count.update()
        else:
            g.observe.start.disable()

        returnValue(status)

    def freeze(self):
        """
        Freeze all settings so they cannot be altered
        """
        self.app.disable()
        self.clear.disable()
        self.nod.disable()
        self.led.disable()
        self.dummy.disable()
        self.readSpeed.disable()
        self.expose.disable()
        self.number.disable()
        self.wframe.disable(everything=True)
        self.nmult.disable()
        self.frozen = True

    def unfreeze(self):
        """
        Reverse of freeze
        """
        self.app.enable()
        self.clear.enable()
        self.nod.enable()
        self.led.enable()
        self.dummy.enable()
        self.readSpeed.enable()
        self.expose.enable()
        self.number.enable()
        self.wframe.enable()
        self.nmult.enable()
        self.frozen = False

    def getRtplotWins(self):
        """ "
        Returns a string suitable to sending off to rtplot when
        it asks for window parameters. Returns null string '' if
        the windows are not OK. This operates on the basis of
        trying to send something back, even if it might not be
        OK as a window setup. Note that we have to take care
        here not to update any GUI components because this is
        called outside of the main thread.
        """
        try:
            if self.isFF():
                return "fullframe\r\n"
            elif self.isDrift():
                xbin = self.wframe.xbin.value()
                ybin = self.wframe.ybin.value()
                nwin = 2 * self.wframe.npair.value()
                ret = str(xbin) + " " + str(ybin) + " " + str(nwin) + "\r\n"
                for xsl, xsr, ys, nx, ny in self.wframe:
                    ret += "{:d} {:d} {:d} {:d}\r\n".format(xsl, ys, nx, ny)
                    ret += "{:d} {:d} {:d} {:d}".format(xsr, ys, nx, ny)
                return ret
            else:
                xbin = self.wframe.xbin.value()
                ybin = self.wframe.ybin.value()
                nwin = 4 * self.wframe.nquad.value()
                ret = str(xbin) + " " + str(ybin) + " " + str(nwin) + "\r\n"
                for xsll, xsul, xslr, xsur, ys, nx, ny in self.wframe:
                    ret += "{:d} {:d} {:d} {:d}\r\n".format(xsll, ys, nx, ny)
                    ret += "{:d} {:d} {:d} {:d}\r\n".format(
                        xsul, 1025 - ys - ny, nx, ny
                    )
                    ret += "{:d} {:d} {:d} {:d}\r\n".format(xslr, ys, nx, ny)
                    ret += "{:d} {:d} {:d} {:d}\r\n".format(
                        xsur, 1025 - ys - ny, nx, ny
                    )
                return ret
        except Exception:
            return ""

    def timing(self):
        """
        Estimates timing information for the current setup. You should
        run a check on the instrument parameters before calling this.

        Returns: (expTime, deadTime, cycleTime, dutyCycle)

        expTime   : exposure time per frame (seconds)
        deadTime  : dead time per frame (seconds)
        cycleTime : sampling time (cadence), (seconds)
        dutyCycle : percentage time exposing.
        frameRate : number of frames per second
        """
        # drift mode y/n?
        isDriftMode = self.isDrift()
        # FF y/n?
        isFF = self.isFF()

        # Set the readout speed
        readSpeed = self.readSpeed()

        if readSpeed == "Fast" and self.dummy():
            video = VIDEO_FAST
        elif readSpeed == "Slow" and self.dummy():
            video = VIDEO_SLOW
        elif not self.dummy():
            video = VIDEO_SLOW_SE
        else:
            raise DriverError(
                "InstPars.timing: readout speed = " + readSpeed + " not recognised."
            )

        if self.fastClk():
            DUMP_TIME = DUMP_TIME_FAST
            VCLOCK_FRAME = VCLOCK_FAST
            VCLOCK_STORAGE = VCLOCK_FAST
            HCLOCK = HCLOCK_FAST
        else:
            DUMP_TIME = DUMP_TIME_SLOW
            VCLOCK_FRAME = VCLOCK_FRAME_SLOW
            VCLOCK_STORAGE = VCLOCK_STORAGE_SLOW
            HCLOCK = HCLOCK_SLOW

        # clear chip on/off?
        lclear = not isDriftMode and self.clear()

        # overscan read or not
        oscan = not isDriftMode and self.oscan()
        oscany = not isDriftMode and self.oscany()

        # get exposure delay
        expose = self.expose.value()

        # window parameters
        xbin = self.wframe.xbin.value()
        ybin = self.wframe.ybin.value()
        if isDriftMode:
            nwin = 1  # number of windows per output
            dys = self.wframe.ys[0].value() - 1
            dnx = self.wframe.nx[0].value()
            dny = self.wframe.ny[0].value()
            dxsl = self.wframe.xsl[0].value()
            dxsr = self.wframe.xsr[0].value()
            # differential shift needed to line both
            # windows up with the edge of the chip
            diffshift = abs(dxsl - 1 - (2 * FFX - dxsr - dnx + 1))
        elif isFF:
            nwin = 1
            ys, nx, ny = [0], [1024], [512]
        else:
            ys, nx, ny = [], [], []
            xse, xsf, xsg, xsh = [], [], [], []
            nwin = self.wframe.nquad.value()
            for xsll, xsul, xslr, xsur, ysv, nxv, nyv in self.wframe:
                xse.append(xsll - 1)
                xsf.append(2049 - xslr - nxv)
                xsg.append(2049 - xsur - nxv)
                xsh.append(xsul - 1)
                ys.append(ysv - 1)
                nx.append(nxv)
                ny.append(nyv)

        # convert timing parameters to seconds
        expose_delay = expose

        # clear chip by VCLOCK-ing the image and area and dumping storage area (x5)
        if lclear:
            clear_time = 5 * (FFY * VCLOCK_FRAME + FFY * DUMP_TIME)
        else:
            clear_time = 0.0

        if isDriftMode:
            # for drift mode, we need the number of windows in the pipeline
            # and the pipeshift
            nrows = FFY  # number of rows in storage area
            pnwin = int(((nrows / dny) + 1) / 2)
            pshift = nrows - (2 * pnwin - 1) * dny
            frame_transfer = (dny + dys) * VCLOCK_FRAME

            yshift = [dys * VCLOCK_STORAGE]

            # After placing the window adjacent to the serial register, the
            # register must be cleared by clocking out the entire register,
            # taking FFX hclocks.
            line_clear = [0.0]
            if yshift[0] != 0:
                line_clear[0] = DUMP_TIME

            # to calculate number of HCLOCKS needed to read a line in
            # drift mode we have to account for the diff shifts and dumping.
            # first perform diff shifts
            # for now we need this *2 (for quadrants E, H or F, G)
            numhclocks = 2 * diffshift
            # now add the amount of clocks needed to get
            # both windows to edge of chip
            if dxsl - 1 > 2 * FFX - dxsr - dnx + 1:
                # it was the left window that got the diff shift,
                # so the number of hclocks increases by the amount
                # needed to get the RH window to the edge
                numhclocks += 2 * FFX - dxsr - dnx + 1
            else:
                # vice versa
                numhclocks += dxsl - 1
            # now we actually clock the windows themselves
            numhclocks += dnx
            # finally, we need to hclock the additional pre-scan pixels
            numhclocks += 2 * PRSCX

            # here is the total time to read the whole line
            line_read = [
                VCLOCK_STORAGE * ybin
                + numhclocks * HCLOCK
                + video * dnx / xbin
                + DUMP_TIME
                + 2 * SETUP_READ
            ]

            readout = [(dny / ybin) * line_read[0]]
        elif isFF:
            # move entire image into storage area
            frame_transfer = FFY * VCLOCK_FRAME + DUMP_TIME

            yshift = [0]
            line_clear = [0]

            numhclocks = FFX + PRSCX
            line_read = [
                VCLOCK_STORAGE * ybin
                + numhclocks * HCLOCK
                + video * nx[0] / xbin
                + SETUP_READ
            ]
            if oscan:
                line_read[0] += video * PRSCX / xbin
            nlines = ny[0] / ybin if not oscany else (ny[0] + 8 / ybin)
            readout = [nlines * line_read[0]]
        else:
            # windowed mode
            # move entire image into storage area
            frame_transfer = FFY * VCLOCK_FRAME + DUMP_TIME

            # dump rows in storage area up to start of the window without changing the
            # image area.
            yshift = nwin * [0.0]
            yshift[0] = ys[0] * DUMP_TIME
            for nw in range(1, nwin):
                yshift[nw] = (ys[nw] - ys[nw - 1] - ny[nw - 1]) * DUMP_TIME

            line_clear = nwin * [0.0]
            # Naidu always dumps the serial register, in windowed mode
            # regardless of whether we need to or not
            for nw in range(nwin):
                line_clear[nw] = DUMP_TIME

            # calculate how long it takes to shift one row into the serial
            # register shift along serial register and then read out the data.
            # total number of hclocks needs to account for diff shifts of
            # windows, carried out in serial
            numhclocks = nwin * [0]
            for nw in range(nwin):
                common_shift = min(xse[nw], xsf[nw], xsg[nw], xsh[nw])
                diffshifts = sum(
                    (xs - common_shift for xs in (xse[nw], xsf[nw], xsg[nw], xsh[nw]))
                )
                numhclocks[nw] = 2 * PRSCX + common_shift + diffshifts + nx[nw]

            line_read = nwin * [0.0]
            # line read includes vclocking a row, all the hclocks, digitising pixels and dumping serial register
            # when windows are read out.
            for nw in range(nwin):
                line_read[nw] = (
                    VCLOCK_STORAGE * ybin
                    + numhclocks[nw] * HCLOCK
                    + video * nx[nw] / xbin
                    + 2 * SETUP_READ
                    + DUMP_TIME
                )
                if oscan:
                    line_read[nw] += video * PRSCX / xbin

            # multiply time to shift one row into serial register by
            # number of rows for total readout time
            readout = nwin * [0.0]
            for nw in range(nwin):
                nlines = ny[nw] / ybin if not oscany else (ny[nw] + 8 / ybin)
                readout[nw] = nlines * line_read[nw]

        # now get the total time to read out one exposure.
        cycleTime = expose_delay + clear_time + frame_transfer
        if isDriftMode:
            cycleTime += (
                pshift * VCLOCK_STORAGE + yshift[0] + line_clear[0] + readout[0]
            )
        else:
            for nw in range(nwin):
                cycleTime += yshift[nw] + line_clear[nw] + readout[nw]

        # use 5sec estimate for nod time
        # TODO: replace with accurate estimate
        if self.nod() and lclear:
            cycleTime += 5
        elif self.nod():
            g = get_root(self).globals
            g.clog.warn("ERR: dithering enabled with clear mode off")

        frameRate = 1.0 / cycleTime
        expTime = expose_delay if lclear else cycleTime - frame_transfer
        deadTime = cycleTime - expTime
        dutyCycle = 100.0 * expTime / cycleTime
        return (expTime, deadTime, cycleTime, dutyCycle, frameRate)


class RunPars(tk.LabelFrame):
    """
    Run parameters
    """

    def __init__(self, master):
        tk.LabelFrame.__init__(
            self, master, text="Next run parameters", padx=10, pady=10
        )

        row = 0
        column = 0
        tk.Label(self, text="Target name").grid(row=row, column=column, sticky=tk.W)

        row += 1
        tk.Label(self, text="Filters").grid(row=row, column=column, sticky=tk.W)

        row += 1
        tk.Label(self, text="Programme ID/OB").grid(row=row, column=column, sticky=tk.W)

        row += 1
        tk.Label(self, text="Principal Investigator").grid(
            row=row, column=column, sticky=tk.W
        )

        row += 1
        tk.Label(self, text="Observer(s)").grid(row=row, column=column, sticky=tk.W)

        row += 1
        tk.Label(self, text="Pre-run comment").grid(row=row, column=column, sticky=tk.W)

        # spacer
        column += 1
        tk.Label(self, text=" ").grid(row=0, column=column)

        # target
        row = 0
        column += 1
        self.target = w.Target(self, self.check)
        self.target.grid(row=row, column=column, sticky=tk.W)

        # filter
        row += 1
        self.filter = w.TextEntry(self, 20, self.check)
        self.filter.grid(row=row, column=column, sticky=tk.W)

        # programme ID / OBID
        row += 1
        self.prog_ob = w.ProgramID(self)
        self.prog_ob.grid(row=row, column=column, sticky=tk.W)

        # principal investigator
        row += 1
        self.pi = w.TextEntry(self, 20, self.check)
        self.pi.grid(row=row, column=column, sticky=tk.W)

        # observers
        row += 1
        self.observers = w.TextEntry(self, 20, self.check)
        self.observers.grid(row=row, column=column, sticky=tk.W)

        # comment
        row += 1
        self.comment = w.TextEntry(self, 38)
        self.comment.grid(row=row, column=column, sticky=tk.W)

    def loadJSON(self, json_string):
        """
        Sets the values of the run parameters given an JSON string
        """
        g = get_root(self).globals
        user = json.loads(json_string)["user"]

        def setField(widget, field):
            val = user.get(field)
            if val is not None:
                widget.set(val.strip())

        setField(self.prog_ob.obid, "OB")
        setField(self.target, "target")
        setField(self.prog_ob.progid, "ID")
        setField(self.pi, "PI")
        setField(self.observers, "Observers")
        setField(self.comment, "comment")
        setField(self.filter, "filters")
        setField(g.observe.rtype, "flags")

    def dumpJSON(self):
        """
        Encodes current parameters to JSON compatible dictionary
        """
        g = get_root(self).globals
        dtype = g.observe.rtype()
        if dtype == "bias":
            target = "BIAS"
        elif dtype == "flat":
            target = "FLAT"
        elif dtype == "dark":
            target = "DARK"
        else:
            target = self.target.value()

        return dict(
            target=target,
            ID=self.prog_ob.progid.value(),
            PI=self.pi.value(),
            OB="{:04d}".format(self.prog_ob.obid.value()),
            Observers=self.observers.value(),
            comment=self.comment.value(),
            flags=dtype,
            filters=self.filter.value(),
        )

    def check(self, *args):
        """
        Checks the validity of the run parameters. Returns
        flag (True = OK), and a message which indicates the
        nature of the problem if the flag is False.
        """

        ok = True
        msg = ""
        g = get_root(self).globals
        dtype = g.observe.rtype()
        expert = g.cpars["expert_level"] > 0

        if dtype == "bias" or dtype == "flat" or dtype == "dark":
            self.pi.configure(state="disable")
            self.prog_ob.configure(state="disable")
            self.target.disable()
        else:
            if expert:
                self.pi.configure(state="normal")
                self.prog_ob.configure(state="normal")
                self.prog_ob.enable()
            else:
                self.prog_ob.configure(state="disable")
                self.pi.configure(state="disable")
                self.prog_ob.disable()
            self.target.enable()

        if g.cpars["require_run_params"]:
            if self.target.ok():
                self.target.entry.config(bg=g.COL["main"])
            else:
                self.target.entry.config(bg=g.COL["error"])
                ok = False
                msg += "Target name field cannot be blank\n"

            if dtype == "acquisition" or dtype == "data" or dtype == "technical":
                if self.prog_ob.ok():
                    self.prog_ob.config(bg=g.COL["main"])
                else:
                    self.prog_ob.config(bg=g.COL["error"])
                    ok = False
                    msg += "Programme or OB ID field cannot be blank\n"

                if self.pi.ok():
                    self.pi.config(bg=g.COL["main"])
                else:
                    self.pi.config(bg=g.COL["error"])
                    ok = False
                    msg += "Principal Investigator field cannot be blank\n"

            if self.observers.ok():
                self.observers.config(bg=g.COL["main"])
            else:
                self.observers.config(bg=g.COL["error"])
                ok = False
                msg += "Observers field cannot be blank"
        return (ok, msg)

    def setExpertLevel(self):
        g = get_root(self).globals
        expert = g.cpars["expert_level"] > 0
        if expert:
            self.pi.configure(state="normal")
            self.prog_ob.configure(state="normal")
            self.prog_ob.enable()
        else:
            self.prog_ob.configure(state="disable")
            self.pi.configure(state="disable")
            self.prog_ob.disable()

    def freeze(self):
        """
        Freeze all settings so that they can't be altered
        """
        self.target.disable()
        self.filter.configure(state="disable")
        self.prog_ob.configure(state="disable")
        self.pi.configure(state="disable")
        self.observers.configure(state="disable")
        self.comment.configure(state="disable")

    def unfreeze(self):
        """
        Unfreeze all settings so that they can be altered
        """
        g = get_root(self).globals
        self.filter.configure(state="normal")
        dtype = g.observe.rtype()
        if dtype == "acquisition" or dtype == "data" or dtype == "technical":
            self.prog_ob.configure(state="normal")
            self.pi.configure(state="normal")
            self.target.enable()
        self.observers.configure(state="normal")
        self.comment.configure(state="normal")


class CountsFrame(tk.LabelFrame):
    """
    Frame for count rate estimates
    """

    def __init__(self, master):
        """
        master : enclosing widget
        """
        tk.LabelFrame.__init__(self, master, pady=2, text="Count & S-to-N estimator")

        # divide into left and right frames
        lframe = tk.Frame(self, padx=2)
        rframe = tk.Frame(self, padx=2)

        # entries
        self.filter = w.Radio(
            lframe, ("u", "g", "r", "i", "z"), 3, self.checkUpdate, initial=1
        )
        self.mag = w.RangedFloat(
            lframe, 18.0, 0.0, 30.0, self.checkUpdate, True, width=5, nplaces=2
        )
        self.seeing = w.RangedFloat(
            lframe, 1.0, 0.2, 20.0, self.checkUpdate, True, True, width=5, nplaces=1
        )
        self.airmass = w.RangedFloat(
            lframe, 1.5, 1.0, 5.0, self.checkUpdate, True, width=5, nplaces=2
        )
        self.moon = w.Radio(lframe, ("d", "g", "b"), 3, self.checkUpdate)

        # results
        self.cadence = w.Ilabel(rframe, text="UNDEF", width=10, anchor=tk.W)
        self.exposure = w.Ilabel(rframe, text="UNDEF", width=10, anchor=tk.W)
        self.duty = w.Ilabel(rframe, text="UNDEF", width=10, anchor=tk.W)
        self.peak = w.Ilabel(rframe, text="UNDEF", width=10, anchor=tk.W)
        self.total = w.Ilabel(rframe, text="UNDEF", width=10, anchor=tk.W)
        self.ston = w.Ilabel(rframe, text="UNDEF", width=10, anchor=tk.W)
        self.ston3 = w.Ilabel(rframe, text="UNDEF", width=10, anchor=tk.W)

        # layout
        # left
        tk.Label(lframe, text="Filter:").grid(
            row=0, column=0, padx=5, pady=3, sticky=tk.W + tk.N
        )
        self.filter.grid(row=0, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text="Mag:").grid(row=1, column=0, padx=5, pady=3, sticky=tk.W)
        self.mag.grid(row=1, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text="Seeing:").grid(
            row=2, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.seeing.grid(row=2, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text="Airmass:").grid(
            row=3, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.airmass.grid(row=3, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(lframe, text="Moon:").grid(
            row=4, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.moon.grid(row=4, column=1, padx=5, pady=3, sticky=tk.W)

        # right
        tk.Label(rframe, text="Cadence:").grid(
            row=0, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.cadence.grid(row=0, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text="Exposure:").grid(
            row=1, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.exposure.grid(row=1, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text="Duty cycle:").grid(
            row=2, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.duty.grid(row=2, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text="Peak:").grid(
            row=3, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.peak.grid(row=3, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text="Total:").grid(
            row=4, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.total.grid(row=4, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text="S/N:").grid(row=5, column=0, padx=5, pady=3, sticky=tk.W)
        self.ston.grid(row=5, column=1, padx=5, pady=3, sticky=tk.W)

        tk.Label(rframe, text="S/N (3h):").grid(
            row=6, column=0, padx=5, pady=3, sticky=tk.W
        )
        self.ston3.grid(row=6, column=1, padx=5, pady=3, sticky=tk.W)

        # slot frames in
        lframe.grid(row=0, column=0, sticky=tk.W + tk.N)
        rframe.grid(row=0, column=1, sticky=tk.W + tk.N)

    def checkUpdate(self, *args):
        """
        Updates values after first checking instrument parameters are OK.
        This is not integrated within update to prevent ifinite recursion
        since update gets called from ipars.
        """
        g = get_root(self).globals
        if not self.check():
            g.clog.warn("Current observing parameters are not valid.")
            return False

        if not g.ipars.check():
            g.clog.warn("Current instrument parameters are not valid.")
            return False

    def check(self):
        """
        Checks values
        """
        status = True
        g = get_root(self).globals
        if self.mag.ok():
            self.mag.config(bg=g.COL["main"])
        else:
            self.mag.config(bg=g.COL["warn"])
            status = False

        if self.airmass.ok():
            self.airmass.config(bg=g.COL["main"])
        else:
            self.airmass.config(bg=g.COL["warn"])
            status = False

        if self.seeing.ok():
            self.seeing.config(bg=g.COL["main"])
        else:
            self.seeing.config(bg=g.COL["warn"])
            status = False

        return status

    def update(self, *args):
        """
        Updates values. You should run a check on the instrument and
        target parameters before calling this.
        """
        g = get_root(self).globals
        expTime, deadTime, cycleTime, dutyCycle, frameRate = g.ipars.timing()

        total, peak, peakSat, peakWarn, ston, ston3 = self.counts(expTime, cycleTime)

        if cycleTime < 0.01:
            self.cadence.config(text="{0:7.5f} s".format(cycleTime))
        elif cycleTime < 0.1:
            self.cadence.config(text="{0:6.4f} s".format(cycleTime))
        elif cycleTime < 1.0:
            self.cadence.config(text="{0:5.3f} s".format(cycleTime))
        elif cycleTime < 10.0:
            self.cadence.config(text="{0:4.2f} s".format(cycleTime))
        elif cycleTime < 100.0:
            self.cadence.config(text="{0:4.1f} s".format(cycleTime))
        elif cycleTime < 1000.0:
            self.cadence.config(text="{0:4.0f} s".format(cycleTime))
        else:
            self.cadence.config(text="{0:5.0f} s".format(cycleTime))

        if expTime < 0.01:
            self.exposure.config(text="{0:7.5f} s".format(expTime))
        elif expTime < 0.1:
            self.exposure.config(text="{0:6.4f} s".format(expTime))
        elif expTime < 1.0:
            self.exposure.config(text="{0:5.3f} s".format(expTime))
        elif expTime < 10.0:
            self.exposure.config(text="{0:4.2f} s".format(expTime))
        elif expTime < 100.0:
            self.exposure.config(text="{0:4.1f} s".format(expTime))
        elif expTime < 1000.0:
            self.exposure.config(text="{0:4.0f} s".format(expTime))
        else:
            self.exposure.config(text="{0:5.0f} s".format(expTime))

        self.duty.config(text="{0:4.1f} %".format(dutyCycle))
        self.peak.config(text="{0:d} cts".format(int(round(peak))))
        if peakSat:
            self.peak.config(bg=g.COL["error"])
        elif peakWarn:
            self.peak.config(bg=g.COL["warn"])
        else:
            self.peak.config(bg=g.COL["main"])

        self.total.config(text="{0:d} cts".format(int(round(total))))
        self.ston.config(text="{0:.1f}".format(ston))
        self.ston3.config(text="{0:.1f}".format(ston3))

    def counts(self, expTime, cycleTime, ap_scale=1.6, ndiv=5):
        """
        Computes counts per pixel, total counts, sky counts
        etc given current magnitude, seeing etc. You should
        run a check on the instrument parameters before calling
        this.

        expTime   : exposure time per frame (seconds)
        cycleTime : sampling, cadence (seconds)
        ap_scale  : aperture radius as multiple of seeing

        Returns: (total, peak, peakSat, peakWarn, ston, ston3)

        total    -- total number of object counts in aperture
        peak     -- peak counts in a pixel
        peakSat  -- flag to indicate saturation
        peakWarn -- flag to indication level approaching saturation
        ston     -- signal-to-noise per exposure
        ston3    -- signal-to-noise after 3 hours on target
        """

        # code directly translated from Java equivalent.
        g = get_root(self).globals

        # Set the readout speed
        readSpeed = g.ipars.readSpeed()
        if readSpeed == "Fast":
            gain = GAIN_FAST
            read = RNO_FAST
        elif readSpeed == "Slow":
            gain = GAIN_SLOW
            read = RNO_SLOW
        else:
            raise DriverError(
                "CountsFrame.counts: readout speed = " + readSpeed + " not recognised."
            )

        xbin, ybin = g.ipars.wframe.xbin.value(), g.ipars.wframe.ybin.value()

        # calculate SN info.
        zero, sky, skyTot, darkTot = 0.0, 0.0, 0.0, 0.0
        total, peak, correct, signal, readTot, seeing = 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
        noise, narcsec, npix, signalToNoise3 = 1.0, 0.0, 0.0, 0.0

        tinfo = g.TINS[g.cpars["telins_name"]]
        filtnam = self.filter.value()

        zero = tinfo["zerop"][filtnam]
        mag = self.mag.value()
        seeing = self.seeing.value()
        sky = g.SKY[self.moon.value()][filtnam]
        airmass = self.airmass.value()
        plateScale = tinfo["plateScale"]

        # calculate expected electrons
        total = 10.0 ** ((zero - mag - airmass * g.EXTINCTION[filtnam]) / 2.5) * expTime

        # compute fraction that fall in central pixel
        # assuming target exactly at its centre. Do this
        # by splitting each pixel of the central (potentially
        # binned) pixel into ndiv * ndiv points at
        # which the seeing profile is added. sigma is the
        # RMS seeing in terms of pixels.
        sigma = seeing / g.EFAC / plateScale

        sum = 0.0
        for iyp in range(ybin):
            yoff = -ybin / 2.0 + iyp
            for ixp in range(xbin):
                xoff = -xbin / 2.0 + ixp
                for iys in range(ndiv):
                    y = (yoff + (iys + 0.5) / ndiv) / sigma
                    for ixs in range(ndiv):
                        x = (xoff + (ixs + 0.5) / ndiv) / sigma
                        sum += math.exp(-(x * x + y * y) / 2.0)
        peak = total * sum / (2.0 * math.pi * sigma**2 * ndiv**2)

        # Work out fraction of flux in aperture with radius AP_SCALE*seeing
        correct = 1.0 - math.exp(-((g.EFAC * ap_scale) ** 2) / 2.0)

        # expected sky e- per arcsec
        skyPerArcsec = 10.0 ** ((zero - sky) / 2.5) * expTime
        # skyPerPixel = skyPerArcsec*plateScale**2*xbin*ybin
        narcsec = math.pi * (ap_scale * seeing) ** 2
        skyTot = skyPerArcsec * narcsec
        npix = math.pi * (ap_scale * seeing / plateScale) ** 2 / xbin / ybin

        signal = correct * total  # in electrons
        darkTot = npix * DARK_E * expTime  # in electrons
        readTot = npix * read**2  # in electrons

        # noise, in electrons
        noise = math.sqrt(readTot + darkTot + skyTot + signal)

        # Now compute signal-to-noise in 3 hour seconds run
        signalToNoise3 = signal / noise * math.sqrt(3 * 3600.0 / cycleTime)

        # convert from electrons to counts
        total /= gain
        peak /= gain

        warn = 25000
        sat = 60000

        peakSat = peak > sat
        peakWarn = peak > warn

        return (total, peak, peakSat, peakWarn, signal / noise, signalToNoise3)


class RunType(w.Select):
    """
    Dropdown box to select run type.

    Start button should be disabled until an option is made from this dropdown.
    """

    DTYPES = ("", "data", "acquire", "bias", "flat", "dark", "tech")
    DVALS = ("", "data", "acquisition", "bias", "flat", "dark", "technical")

    def __init__(self, master, start_button, checker=None):
        w.Select.__init__(self, master, 0, RunType.DTYPES, self.check)
        self.start_button = start_button
        self._checker = checker

    def __call__(self):
        index = self.options.index(self.val.get())
        return RunType.DVALS[index]

    def set(self, value):
        index = RunType.DVALS.index(value)
        w.Select.set(self, RunType.DTYPES[index])

    @inlineCallbacks
    def check(self, *args):
        if self._checker is not None:
            self._checker()
        if self.val.get() == "":
            self.start_button.run_type_set = False
            self.start_button.disable()
        else:
            self.start_button.run_type_set = True
            g = get_root(self).globals
            try:
                run_active = yield isRunActive(g)
                powered_on = yield isPoweredOn(g)
            except Exception as err:
                g.clog.warn(str(err))
            if (
                g.cpars["hcam_server_on"]
                and g.cpars["eso_server_online"]
                and g.observe.start["state"] == "disabled"
                and not run_active
                and powered_on
            ):
                self.start_button.enable()
            g.rpars.check()


class Start(w.ActButton):
    """
    Button to start a run.

    This is a complex process including the following steps:

    -- check the instrument and run parameters are OK
    -- optionally, hassle the user if the target changes
    -- post the run settings to the ESO NGC control server
    -- start the run
    -- update the button status
    -- start the exposure timer
    """

    def __init__(self, master, width):
        """
        Parameters
        ----------
        master : tk
            containing widget
        width : int
            width of button
        """
        w.ActButton.__init__(self, master, width, text="Start")
        g = get_root(self).globals
        self.config(bg=g.COL["start"])
        self.target = None
        self.run_type_set = False

    def enable(self):
        """
        Enable the button
        """
        if self.run_type_set:
            w.ActButton.enable(self)
            g = get_root(self).globals
            self.config(bg=g.COL["start"])

    def disable(self):
        """
        Disable the button, if in non-expert mode.
        """
        w.ActButton.disable(self)
        g = get_root(self).globals
        if self._expert:
            self.config(bg=g.COL["start"])
        else:
            self.config(bg=g.COL["startD"])

    def setExpert(self):
        """
        Turns on 'expert' status whereby the button is always enabled,
        regardless of its activity status.
        """
        w.ActButton.setExpert(self)
        g = get_root(self).globals
        self.config(bg=g.COL["start"])

    def setNonExpert(self):
        """
        Turns off 'expert' status whereby to allow a button to be disabled
        """
        self._expert = False
        if self._active and self.run_type_set:
            self.enable()
        else:
            self.disable()

    @inlineCallbacks
    def on_telemetry(self, package):
        """
        This is run every time a telemetry packet comes in from NGC.

        It is the responsibility of an implementing GUI to subscribe to the
        NGC telemetry topic with this function as the callback.
        """
        telemetry = pickle.loads(package)
        res = ReadNGCTelemetry(telemetry)
        if not res.ok:
            raise DriverError("cannot read NGC telemetry: " + str(res.err))
        if res.clocks != "enabled":
            # NGC voltages are not powered on, cannot start runs
            self.disable()
        elif res.state == "active":
            # run is underway - cannot start runs
            self.disable()
        else:
            self.enable()

    @inlineCallbacks
    def act(self):
        """
        Carries out action associated with start button
        """
        g = get_root(self).globals
        # check binning against overscan
        msg = """
        HiperCAM has an o/scan of 50 pixels.
        Your binning does not fit into this
        region. Some columns will contain a
        mix of o/scan and data.

        Click OK if you wish to continue."""
        if g.ipars.oscan():
            xbin, ybin = g.ipars.wframe.xbin.value(), g.ipars.wframe.ybin.value()
            if xbin not in (1, 2, 5, 10) or ybin not in (1, 2, 5, 10):
                if not messagebox.askokcancel("Binning alert", msg):
                    returnValue(False)

        # Check instrument pars are OK
        if not g.ipars.check():
            g.clog.warn("Invalid instrument parameters; start failed.")
            returnValue(False)

        # create JSON to post
        data = yield createJSON(g)

        # check if COMPO is in position
        # Do this regardless if enabled or not, as we might need to park
        if not g.compo_hw.ok_to_start_run:
            msg = """
            COMPO is reporting that it is not ready to start a run.
            Please check the state of COMPO.
            
            Click OK if you wish to continue anyway."""
            if not messagebox.askokcancel("COMPO alert", msg):
                returnValue(False)

        # check autoguiding is started if we are guiding with COMPO
        if g.ipars.compo() and g.compo_hw.setup_frame.injection_side.value() == "G":
            msg = """
            COMPO setup implies you will be guiding with COMPO.
            Check that autoguiding is set up and running.
            
            Click OK when you wish to continue and start run.
            Click Cancel to abort run."""
            if not messagebox.askokcancel("Guiding alert", msg):
                returnValue(False)

        # check time to limit for rotator
        if (
            g.info.time_to_limit is not None
            and g.info.time_to_limit < g.info.WARN_LIMIT
        ):
            msg = f"""
            The rotator limit will be reach in less than {g.info.WARN_LIMIT.value} hour.

            You may want to de-rotate before starting the run. 
            Click OK when you wish to continue and start run.
            Click Cancel to abort run."""
            if not messagebox.askokcancel("Rotator alert", msg):
                returnValue(False)

        # POST
        try:
            success = yield postJSON(g, data)
            if not success:
                raise Exception("postJSON returned False")
        except Exception as err:
            g.clog.warn("Failed to post data to servers")
            g.clog.warn(str(err))
            returnValue(False)

        # START
        try:
            success = yield execCommand(g, "start")
            if not success:
                raise Exception("Start command failed: check server response")
        except Exception as err:
            g.clog.warn("Failed to start run")
            g.clog.warn(str(err))
            returnValue(False)

        # Is nod enabled? Should we start GTC offsetter?
        try:
            success = yield startNodding(g, data)
            if not success:
                raise Exception("Failed to start dither: response was false")
        except Exception as err:
            g.clog.warn("Failed to start GTC offsetter")
            g.clog.warn(str(err))
            g.clog.warn("Run may be paused indefinitely")
            g.clog.warn('use "ngcbCmd seq start" to fix')
            returnValue(False)

        # Run successfully started.
        # enable stop button, disable Start
        # also make inactive until RunType select box makes active again
        # start run timer
        # finally, clear table which stores TCS info during this run
        self.disable()
        self.run_type_set = False
        g.observe.stop.enable()
        g.info.timer.start()
        g.info.clear_tcs_table()
        returnValue(True)


class Load(w.ActButton):
    """
    Class defining the 'Load' button's operation. This loads a previously
    saved configuration from disk.
    """

    def __init__(self, master, width):
        """
        master  : containing widget
        width   : width of button
        """
        w.ActButton.__init__(self, master, width, text="Load")

    def act(self):
        """
        Carries out the action associated with the Load button
        """
        g = get_root(self).globals
        fname = filedialog.askopenfilename(
            defaultextension=".json",
            filetypes=[("json files", ".json"), ("fits files", ".fits")],
            initialdir=g.cpars["app_directory"],
        )
        if not fname:
            g.clog.warn("Aborted load from disk")
            return False

        # load json
        if fname.endswith(".json"):
            with open(fname) as ifname:
                json_string = ifname.read()
        else:
            json_string = jsonFromFits(fname)

        # load up the instrument settings
        g.ipars.loadJSON(json_string)

        # load up the run parameters
        g.rpars.loadJSON(json_string)

        # load the COMPO setup
        g.compo_hw.loadJSON(json_string)

        return True


class Save(w.ActButton):
    """
    Class defining the 'Save' button's operation. This saves the
    current configuration to disk.
    """

    def __init__(self, master, width):
        """
        master  : containing widget
        width   : width of button
        """
        w.ActButton.__init__(self, master, width, text="Save")

    @inlineCallbacks
    def act(self):
        """
        Carries out the action associated with the Save button
        """
        g = get_root(self).globals
        g.clog.info("\nSaving current application to disk")

        # check instrument parameters
        if not g.ipars.check():
            g.clog.warn("Invalid instrument parameters; save failed.")
            returnValue(False)

        # check run parameters
        rok, msg = g.rpars.check()
        if not rok:
            g.clog.warn("Invalid run parameters; save failed.")
            g.clog.warn(msg)
            returnValue(False)

        # Get data to save
        data = yield createJSON(g, full=False)

        # Save to disk
        if saveJSON(g, data):
            # modify buttons
            g.observe.load.enable()
            g.observe.unfreeze.disable()

            # unfreeze the instrument and run params
            g.ipars.unfreeze()
            g.rpars.unfreeze()
            returnValue(True)
        else:
            returnValue(False)


class Unfreeze(w.ActButton):
    """
    Class defining the 'Unfreeze' button's operation.
    """

    def __init__(self, master, width):
        """
        master  : containing widget
        width   : width of button
        """
        w.ActButton.__init__(self, master, width, text="Unfreeze")

    def act(self):
        """
        Carries out the action associated with the Unfreeze button
        """
        g = get_root(self).globals
        g.ipars.unfreeze()
        g.rpars.unfreeze()
        g.observe.load.enable()
        self.disable()


class Observe(tk.LabelFrame):
    """
    Observe widget. Collects together all the buttons needed for observing.
    """

    def __init__(self, master):
        tk.LabelFrame.__init__(self, master, padx=10, pady=10)

        width = 10
        self.load = Load(self, width)
        self.save = Save(self, width)
        self.unfreeze = Unfreeze(self, width)
        self.start = Start(self, width)
        self.rtype = RunType(self, self.start)
        self.stop = w.Stop(self, width)

        # Lay them out
        self.load.grid(row=0, column=0)
        self.save.grid(row=1, column=0)
        self.unfreeze.grid(row=2, column=0)
        self.rtype.grid(row=0, column=1, sticky=tk.EW)
        self.start.grid(row=1, column=1)
        self.stop.grid(row=2, column=1)

        # Define initial status
        self.start.disable()
        self.stop.disable()
        self.unfreeze.disable()

        # Implement expert level
        self.setExpertLevel()
        self.telemetry_topics = [("hipercam.ngc.telemetry", self.on_telemetry)]

    def on_telemetry(self, package):
        self.stop.on_telemetry(package)
        self.start.on_telemetry(package)

    def setExpertLevel(self):
        """
        Set expert level
        """
        g = get_root(self).globals
        level = g.cpars["expert_level"]

        # now set whether buttons are permanently enabled or not
        if level == 0 or level == 1:
            self.load.setNonExpert()
            self.save.setNonExpert()
            self.unfreeze.setNonExpert()
            self.start.setNonExpert()
            self.stop.setNonExpert()

        elif level == 2:
            self.load.setExpert()
            self.save.setExpert()
            self.unfreeze.setExpert()
            self.start.setExpert()
            self.stop.setExpert()
